#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2020 Pyresample developers.
#
# This file is part of Pyresample
#
# Author(s):
#
#   Panu Lahtinen <panu.lahtinen@fmi.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Code for resampling using bilinear algorithm for irregular grids.

The algorithm is taken from

http://www.ahinson.com/algorithms_general/Sections/InterpolationRegression/InterpolationIrregularBilinear.pdf

"""

import warnings

import numpy as np
from pykdtree.kdtree import KDTree

from pyresample import data_reduce, geometry
from pyresample._spatial_mp import Proj

from ..future.resamplers._transform_utils import lonlat2xyz


class BilinearBase(object):
    """Base class for bilinear interpolation."""

    def __init__(self,
                 source_geo_def,
                 target_geo_def,
                 radius_of_influence,
                 neighbours=32,
                 epsilon=0,
                 reduce_data=True):
        """
        Initialize resampler.

        Parameters
        ----------
        source_geo_def : object
            Geometry definition of source
        target_geo_def : object
            Geometry definition of target
        radius_of_influence : float
            Cut off distance in meters
        neighbours : int, optional
            The number of neigbours to consider for each grid point
        epsilon : float, optional
            Allowed uncertainty in meters. Increasing uncertainty
            reduces execution time
        reduce_data : bool, optional
            Perform initial coarse reduction of source dataset in order
            to reduce execution time

        """
        self.bilinear_t = None
        self.bilinear_s = None
        self.slices_x = None
        self.slices_y = None
        self.mask_slices = None
        self.out_coords_x = None
        self.out_coords_y = None

        self._out_coords = {'x': self.out_coords_x, 'y': self.out_coords_y}
        self._valid_input_index = None
        self._index_array = None
        self._distance_array = None
        self._neighbours = neighbours
        self._epsilon = epsilon
        self._reduce_data = reduce_data
        self._source_geo_def = source_geo_def
        self._target_geo_def = target_geo_def
        self._radius_of_influence = radius_of_influence
        self._resample_kdtree = None
        self._target_lons = None
        self._target_lats = None

    def resample(self, data, fill_value=None, nprocs=1):
        """Resample the given data."""
        raise NotImplementedError

    def get_bil_info(self, kdtree_class=KDTree, nprocs=1):
        """Calculate bilinear neighbour info."""
        if self._source_geo_def.size < self._neighbours:
            warnings.warn('Searching for %s neighbours in %s data points' %
                          (self._neighbours, self._source_geo_def.size))

        self._get_valid_input_index_and_kdtree(
            kdtree_class=kdtree_class,
            nprocs=nprocs
        )
        if self._resample_kdtree is None:
            return

        self._target_lons, self._target_lats = self._target_geo_def.get_lonlats()
        self._get_index_array()

        # Calculate vertical and horizontal fractional distances t and s
        self._get_fractional_distances()
        self._get_target_proj_vectors()
        self._get_slices()

    def _get_valid_input_index_and_kdtree(self, kdtree_class=KDTree, nprocs=1):
        valid_input_index, resample_kdtree = self._create_resample_kdtree(
            kdtree_class=kdtree_class,
            nprocs=nprocs
        )

        if resample_kdtree:
            self._valid_input_index = valid_input_index
            self._resample_kdtree = resample_kdtree
        else:
            # Handle if all input data is reduced away
            self._create_empty_bil_info()

    def _create_empty_bil_info(self):
        self._valid_input_index = np.ones(self._source_geo_def.size, dtype=bool)
        self._index_array = np.ones((self._target_geo_def.size, 4), dtype=np.int32)
        self.bilinear_s = np.nan * np.zeros(self._target_geo_def.size)
        self.bilinear_t = np.nan * np.zeros(self._target_geo_def.size)
        self.slices_x = np.zeros((self._target_geo_def.size, 4), dtype=np.int32)
        self.slices_y = np.zeros((self._target_geo_def.size, 4), dtype=np.int32)
        self.out_coords_x, self.out_coords_y = self._target_geo_def.get_proj_vectors()
        self.mask_slices = self._index_array >= self._source_geo_def.size

    def _get_valid_output_indices(self):
        return ((self._target_lons >= -180) & (self._target_lons <= 180) &
                (self._target_lats <= 90) & (self._target_lats >= -90))

    def _get_index_array(self):
        index_array = _query_no_distance(
            self._target_lons, self._target_lats,
            self._get_valid_output_indices(), self._resample_kdtree,
            self._neighbours, self._epsilon,
            self._radius_of_influence)
        self._index_array = self._reduce_index_array(index_array)

    def _reduce_index_array(self, index_array):
        input_size = np.sum(self._valid_input_index)
        index_mask = index_array == input_size
        return np.where(index_mask, 0, index_array)

    def _get_fractional_distances(self):
        out_x, out_y = self._get_output_xy()
        # Get the four closest corner points around each output location
        corner_points, self._index_array = \
            _get_four_closest_corners(*self._get_input_xy(),
                                      out_x, out_y,
                                      self._neighbours, self._index_array)
        self.bilinear_t, self.bilinear_s = _get_fractional_distances(
            corner_points, out_x, out_y)

    def _get_output_xy(self):
        return _get_output_xy(self._target_geo_def)

    def _get_input_xy(self):
        return _get_input_xy(self._source_geo_def,
                             Proj(self._target_geo_def.proj_str),
                             self._valid_input_index, self._index_array)

    def _get_target_proj_vectors(self):
        try:
            self.out_coords_x, self.out_coords_y = self._target_geo_def.get_proj_vectors()
        except AttributeError:
            pass

    def _get_slices(self):
        shp = self._source_geo_def.shape
        cols, lines = np.meshgrid(np.arange(shp[1]),
                                  np.arange(shp[0]))

        valid_lines_and_columns = array_slice_for_multiple_arrays(
            self._valid_input_index,
            (np.ravel(lines), np.ravel(cols))
        )
        self.slices_y, self.slices_x = array_slice_for_multiple_arrays(
            self._index_array,
            valid_lines_and_columns
        )
        self.mask_slices = self._index_array >= self._source_geo_def.size

    def _create_resample_kdtree(self, kdtree_class=KDTree, nprocs=1):
        """Set up kd tree on input."""
        valid_input_index, input_coords = self._get_valid_input_index_and_input_coords()
        kdtree = None
        if input_coords.size:
            if nprocs > 1:
                kdtree = kdtree_class(input_coords, nprocs=nprocs)
            else:
                kdtree = KDTree(input_coords)
        return valid_input_index, kdtree

    def _get_valid_input_index_and_input_coords(self):
        valid_input_index, source_lons, source_lats = \
            _get_valid_input_index(self._source_geo_def,
                                   self._target_geo_def,
                                   self._reduce_data,
                                   self._radius_of_influence)
        input_coords = lonlat2xyz(source_lons, source_lats)
        valid_input_index = np.ravel(valid_input_index)
        input_coords = input_coords[valid_input_index, :].astype(np.float64)

        return valid_input_index, input_coords

    def get_sample_from_bil_info(self, data, fill_value=None, output_shape=None):
        """Resample using pre-computed resampling LUTs."""
        del output_shape
        fill_value = _check_fill_value(fill_value, data.dtype)

        res = _resample(
            self._slice_data(data, fill_value),
            (self.bilinear_s, self.bilinear_t)
        )

        return self._finalize_output_data(data, res, fill_value)

    def _slice_data(self, data, fill_value):
        raise NotImplementedError

    def _finalize_output_data(self, data, res, fill_value):
        raise NotImplementedError

    def save_resampling_info(self, filename):
        """Save bilinear resampling look-up tables."""
        raise NotImplementedError

    def load_resampling_info(self, filename):
        """Load bilinear resampling look-up tables and initialize the resampler."""
        raise NotImplementedError


def _check_fill_value(fill_value, dtype):
    """Check that fill value is usable for the data."""
    if fill_value is None:
        if np.issubdtype(dtype, np.integer):
            fill_value = 0
        else:
            fill_value = np.nan
    elif np.issubdtype(dtype, np.integer):
        if np.isnan(fill_value):
            fill_value = 0
        elif np.issubdtype(type(fill_value), np.floating):
            fill_value = int(fill_value)

    return fill_value


def _get_output_xy(target_geo_def):
    out_x, out_y = target_geo_def.get_proj_coords()
    return np.ravel(out_x), np.ravel(out_y)


def _get_input_xy(source_geo_def, proj, valid_input_index, index_array):
    """Get x/y coordinates for the input area and reduce the data."""
    input_xy_coordinates = mask_coordinates(*_get_raveled_lonlats(source_geo_def))
    valid_xy_coordinates = array_slice_for_multiple_arrays(valid_input_index, input_xy_coordinates)
    # Expand input coordinates for each output location
    expanded_coordinates = array_slice_for_multiple_arrays(index_array, valid_xy_coordinates)

    return proj(*expanded_coordinates)


def array_slice_for_multiple_arrays(idxs, data):
    """Slices multiple arrays using the same indices."""
    return [d[idxs] for d in data]


def mask_coordinates(lons, lats):
    """Mask invalid coordinate values."""
    idxs = (find_indices_outside_min_and_max(lons, -180., 180.) |
            find_indices_outside_min_and_max(lats, -90., 90.))
    return _np_where_for_multiple_arrays(idxs, (np.nan, np.nan), (lons, lats))


def find_indices_outside_min_and_max(data, min_val, max_val):
    """Return array indices outside the given minimum and maximum values."""
    return (data < min_val) | (data > max_val)


def _get_raveled_lonlats(geo_def):
    lons, lats = geo_def.get_lonlats()
    if lons.size == 0 or lats.size == 0:
        raise ValueError('Cannot resample empty data set')
    elif lons.size != lats.size or lons.shape != lats.shape:
        raise ValueError('Mismatch between lons and lats')

    return np.ravel(lons), np.ravel(lats)


def _get_four_closest_corners(in_x, in_y, out_x, out_y, neighbours, index_array):
    """Get bounding corners.

    Get four closest locations from (in_x, in_y) so that they form a
    bounding rectangle around the requested location given by (out_x,
    out_y).

    """
    # Find four closest pixels around the target location

    stride, valid_corners = _get_stride_and_valid_corner_indices(
        out_x, out_y, in_x, in_y, neighbours)
    res = []
    indices = []
    for valid in valid_corners:
        x__, y__, idx = _get_corner(stride, valid, in_x, in_y, index_array)
        res.append(np.transpose(np.vstack((x__, y__))))
        indices.append(idx)

    return res, np.transpose(np.vstack(indices))


def _get_fractional_distances(corner_points, out_x, out_y):
    """Calculate vertical and horizontal fractional distances t and s."""
    # General case, ie. where the the corners form an irregular rectangle
    t__, s__ = _invalid_s_and_t_to_nan(
        *_get_fractional_distances_irregular(corner_points, out_y, out_x)
    )
    # Update where verticals are parallel
    t__, s__ = _update_fractional_distances(
        _get_fractional_distances_uprights_parallel,
        t__, s__, corner_points, out_x, out_y
    )
    # Update where both verticals and horizontals are parallel
    corner_points = (corner_points[0], corner_points[1], corner_points[2])
    t__, s__ = _update_fractional_distances(
        _get_fractional_distances_parallellogram,
        t__, s__, corner_points, out_x, out_y
    )

    return t__, s__


def _invalid_s_and_t_to_nan(t__, s__):
    return _np_where_for_multiple_arrays(
        (find_indices_outside_min_and_max(t__, 0, 1) |
         find_indices_outside_min_and_max(s__, 0, 1)),
        (np.nan, np.nan),
        (t__, s__))


def _get_fractional_distances_irregular(corner_points, out_y, out_x):
    """Get parameters for the case where none of the sides are parallel."""
    # Get the valid roots from interval [0, 1]
    t__ = _solve_quadratic(
        *_calc_abc(corner_points, out_y, out_x),
        min_val=0., max_val=1.)

    # Calculate parameter s
    pt_1, pt_2, pt_3, pt_4 = corner_points
    y_corners = (pt_1[:, 1], pt_3[:, 1], pt_2[:, 1], pt_4[:, 1])
    s__ = _solve_another_fractional_distance(t__, y_corners, out_y)

    return t__, s__


def _solve_quadratic(a__, b__, c__, min_val=0.0, max_val=1.0):
    """Solve quadratic equation.

    Solve quadratic equation and return the valid roots from interval
    [*min_val*, *max_val*].

    """
    a__ = _ensure_array(a__)
    b__ = _ensure_array(b__)
    c__ = _ensure_array(c__)
    discriminant = b__ * b__ - 4 * a__ * c__

    # Solve the quadratic polynomial
    x_1 = (-b__ + np.sqrt(discriminant)) / (2 * a__)
    x_2 = (-b__ - np.sqrt(discriminant)) / (2 * a__)
    # Linear case
    x_3 = -c__ / b__

    # Find valid solutions, ie. 0 <= t <= 1
    x__ = np.where(
        find_indices_outside_min_and_max(x_1, min_val, max_val) | np.isnan(x_1),
        x_2, x_1)

    x__ = np.where(
        find_indices_outside_min_and_max(x__, min_val, max_val) | np.isnan(x__),
        x_3, x__)

    x__ = np.where(
        find_indices_outside_min_and_max(x__, min_val, max_val),
        np.nan, x__)

    return x__


def _ensure_array(val):
    """Ensure that *val* is an array-like object."""
    if np.isscalar(val):
        val = np.array([val])
    return val


def _calc_abc(corner_points, out_y, out_x):
    """Calculate coefficients for quadratic equation.

    In this order of arguments used for _get_fractional_distances_irregular() and
    _get_fractional_distances_uprights().  For _get_fractional_distances_uprights switch order of pt_2 and
    pt_3.

    """
    pt_1, pt_2, pt_3, pt_4 = corner_points
    # Pairwise longitudal separations between reference points
    x_21 = pt_2[:, 0] - pt_1[:, 0]
    x_31 = pt_3[:, 0] - pt_1[:, 0]
    x_42 = pt_4[:, 0] - pt_2[:, 0]

    # Pairwise latitudal separations between reference points
    y_21 = pt_2[:, 1] - pt_1[:, 1]
    y_31 = pt_3[:, 1] - pt_1[:, 1]
    y_42 = pt_4[:, 1] - pt_2[:, 1]

    a__ = x_31 * y_42 - y_31 * x_42
    b__ = out_y * (x_42 - x_31) - out_x * (y_42 - y_31) + \
        x_31 * pt_2[:, 1] - y_31 * pt_2[:, 0] + \
        y_42 * pt_1[:, 0] - x_42 * pt_1[:, 1]
    c__ = out_y * x_21 - out_x * y_21 + pt_1[:, 0] * pt_2[:, 1] - \
        pt_2[:, 0] * pt_1[:, 1]

    return a__, b__, c__


def _solve_another_fractional_distance(f__, y_corners, out_y):
    """Solve parameter t__ from s__, or vice versa.

    For solving s__, switch order of y_2 and y_3.
    """
    y_1, y_2, y_3, y_4 = y_corners

    y_21 = y_2 - y_1
    y_43 = y_4 - y_3

    g__ = ((out_y - y_1 - y_21 * f__) /
           (y_3 + y_43 * f__ - y_1 - y_21 * f__))

    # Limit values to interval [0, 1]
    g__ = np.where(
        find_indices_outside_min_and_max(g__, 0, 1),
        np.nan, g__)

    return g__


def _update_fractional_distances(func, t__, s__, corner_points, out_x, out_y):
    idxs = np.ravel(np.isnan(t__) | np.isnan(s__))
    if np.any(idxs):
        new_values = func(corner_points, out_y, out_x)
        updated_t_and_s = _np_where_for_multiple_arrays(idxs, new_values, (t__, s__))
        t__, s__ = _invalid_s_and_t_to_nan(*updated_t_and_s)
    return t__, s__


def _get_fractional_distances_uprights_parallel(corner_points, out_y, out_x):
    """Get parameters for the case where uprights are parallel."""
    pt_1, pt_2, pt_3, pt_4 = corner_points
    # Get the valid roots from interval [0, 1]. Note the different order needed here.
    corner_points = (pt_1, pt_3, pt_2, pt_4)
    s__ = _solve_quadratic(
        *_calc_abc(corner_points, out_y, out_x),
        min_val=0., max_val=1.)

    # Calculate parameter t
    y_corners = (pt_1[:, 1], pt_2[:, 1], pt_3[:, 1], pt_4[:, 1])
    t__ = _solve_another_fractional_distance(s__, y_corners, out_y)

    return t__, s__


def _get_fractional_distances_parallellogram(points, out_y, out_x):
    """Get parameters for the case where uprights are parallel."""
    pt_1, pt_2, pt_3 = points
    # Pairwise longitudal separations between reference points
    x_21 = pt_2[:, 0] - pt_1[:, 0]
    x_31 = pt_3[:, 0] - pt_1[:, 0]

    # Pairwise latitudal separations between reference points
    y_21 = pt_2[:, 1] - pt_1[:, 1]
    y_31 = pt_3[:, 1] - pt_1[:, 1]

    t__ = (x_21 * (out_y - pt_1[:, 1]) - y_21 * (out_x - pt_1[:, 0])) / \
          (x_21 * y_31 - y_21 * x_31)
    t__ = np.where(
        find_indices_outside_min_and_max(t__, 0., 1.),
        np.nan, t__)

    s__ = (out_x - pt_1[:, 0] + x_31 * t__) / x_21
    s__ = np.where(
        find_indices_outside_min_and_max(s__, 0., 1.),
        np.nan, s__)

    return t__, s__


def _get_stride_and_valid_corner_indices(out_x, out_y, in_x, in_y, neighbours):
    out_x_tile = _tile_output_coordinate_vector(out_x, neighbours)
    out_y_tile = _tile_output_coordinate_vector(out_y, neighbours)

    # Get differences in both directions
    x_diff = out_x_tile - in_x
    y_diff = out_y_tile - in_y

    return (np.arange(x_diff.shape[0]), (
        (x_diff > 0) & (y_diff < 0),  # Upper left corners
        (x_diff < 0) & (y_diff < 0),  # Upper right corners
        (x_diff > 0) & (y_diff > 0),  # Lower left corners
        (x_diff < 0) & (y_diff > 0))  # Lower right corners
    )


def _tile_output_coordinate_vector(vector, neighbours):
    return np.reshape(np.tile(vector, neighbours), (neighbours, vector.size)).T


def _get_corner(stride, valid, in_x, in_y, index_array):
    """Get closest set of coordinates from the *valid* locations."""
    x__, y__, idx = _slice_2d_with_stride_and_indices_for_multiple_arrays(
        (in_x, in_y, index_array),
        stride,
        np.argmax(valid, axis=1)  # The closest valid locations
    )
    # Replace invalid points with np.nan
    x__, y__ = _np_where_for_multiple_arrays(
        np.invert(np.max(valid, axis=1)),
        (np.nan, np.nan),
        (x__, y__)
    )

    return x__, y__, idx


def _slice_2d_with_stride_and_indices_for_multiple_arrays(arrays, stride, idxs):
    return [arr[stride, idxs] for arr in arrays]


def _np_where_for_multiple_arrays(idxs, values_for_idxs, otherwise_arrays):
    return [np.where(idxs, values_for_idxs[i], arr) for i, arr in enumerate(otherwise_arrays)]


def _get_valid_input_index(source_geo_def,
                           target_geo_def,
                           reduce_data,
                           radius_of_influence):
    """Find indices of reduced input data."""
    source_lons, source_lats = _get_raveled_lonlats(source_geo_def)

    valid_input_index = np.invert(
        find_indices_outside_min_and_max(source_lons, -180., 180.) |
        find_indices_outside_min_and_max(source_lats, -90., 90.))

    if reduce_data and is_swath_to_grid_or_grid_to_grid(source_geo_def, target_geo_def):
        valid_input_index &= get_valid_indices_from_lonlat_boundaries(
            target_geo_def, source_lons, source_lats, radius_of_influence)

    return valid_input_index, source_lons, source_lats


def is_swath_to_grid_or_grid_to_grid(source_geo_def, target_geo_def):
    """Check whether the resampling is from swath or grid to grid."""
    return (isinstance(source_geo_def, geometry.CoordinateDefinition) and
            isinstance(target_geo_def, (geometry.GridDefinition,
                                        geometry.AreaDefinition))) or \
           (isinstance(source_geo_def, (geometry.GridDefinition,
                                        geometry.AreaDefinition)) and
            isinstance(target_geo_def, (geometry.GridDefinition,
                                        geometry.AreaDefinition)))


def get_valid_indices_from_lonlat_boundaries(
        target_geo_def, source_lons, source_lats, radius_of_influence):
    """Get valid indices from lonlat boundaries."""
    # Resampling from swath to grid or from grid to grid
    lonlat_boundary = target_geo_def.get_boundary_lonlats()

    # Combine reduced and legal values
    return data_reduce.get_valid_index_from_lonlat_boundaries(
        lonlat_boundary[0],
        lonlat_boundary[1],
        source_lons, source_lats,
        radius_of_influence)


def get_slicer(data):
    """Get slicing function for 2D or 3D arrays, depending on the data."""
    if data.ndim == 2:
        return _slice2d
    elif data.ndim == 3:
        return _slice3d
    else:
        raise ValueError("Only 2D and 3D arrays are supported")


def _slice2d(values, sl_x, sl_y, mask, fill_value):
    # Slice 2D data
    arr = values[(sl_y, sl_x)]
    arr[(mask, )] = fill_value
    return arr[:, 0], arr[:, 1], arr[:, 2], arr[:, 3]


def _slice3d(values, sl_x, sl_y, mask, fill_value):
    # Slice 3D data
    arr = values[(slice(None), sl_y, sl_x)]
    arr[(slice(None), mask)] = fill_value
    return arr[:, :, 0], arr[:, :, 1], arr[:, :, 2], arr[:, :, 3]


def _resample(corner_points, fractional_distances):
    p_1, p_2, p_3, p_4 = corner_points
    s__, t__ = fractional_distances
    return (p_1 * (1 - s__) * (1 - t__) +
            p_2 * s__ * (1 - t__) +
            p_3 * (1 - s__) * t__ +
            p_4 * s__ * t__)


def _query_no_distance(target_lons, target_lats,
                       valid_output_index, kdtree, neighbours, epsilon, radius):
    """Query the kdtree. No distances are returned."""
    voir = np.ravel(valid_output_index)
    target_lons_valid = np.ravel(target_lons)[voir]
    target_lats_valid = np.ravel(target_lats)[voir]

    _, index_array = kdtree.query(
        lonlat2xyz(target_lons_valid, target_lats_valid),
        k=neighbours,
        eps=epsilon,
        distance_upper_bound=radius)

    return index_array



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Helper functions related to transforming coordinates."""
import numpy as np


def lonlat2xyz(lons, lats):
    """Convert lon/lat degrees to geocentric x/y/z coordinates."""
    R = 6370997.0
    lats = np.deg2rad(lats)
    r_cos_lats = R * np.cos(lats)
    lons = np.deg2rad(lons)
    x_coords = r_cos_lats * np.cos(lons)
    y_coords = r_cos_lats * np.sin(lons)
    z_coords = R * np.sin(lats)

    return np.stack(
        (x_coords.ravel(), y_coords.ravel(), z_coords.ravel()), axis=-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Utilities for working with rasterio objects."""
from . import proj4_str_to_dict


def _get_area_def_from_gdal(dataset, area_id=None, name=None, proj_id=None, proj_dict=None):
    from pyresample.geometry import AreaDefinition

    # a: width of a pixel
    # b: row rotation (typically zero)
    # c: x-coordinate of the upper-left corner of the upper-left pixel
    # d: column rotation (typically zero)
    # e: height of a pixel (typically negative)
    # f: y-coordinate of the of the upper-left corner of the upper-left pixel
    c, a, b, f, d, e = dataset.GetGeoTransform()
    if not (b == d == 0):
        raise ValueError('Rotated rasters are not supported at this time.')
    area_extent = (c, f + e * dataset.RasterYSize, c + a * dataset.RasterXSize, f)

    if proj_dict is None:
        from osgeo import osr
        proj = dataset.GetProjection()
        if proj != '':
            sref = osr.SpatialReference(wkt=proj)
            proj_dict = proj4_str_to_dict(sref.ExportToProj4())
        else:
            raise ValueError('The source raster is not gereferenced, please provide the value of proj_dict')

        if proj_id is None:
            proj_id = proj.split('"')[1]

    area_def = AreaDefinition(area_id, name, proj_id, proj_dict,
                              dataset.RasterXSize, dataset.RasterYSize, area_extent)
    return area_def


def _get_area_def_from_rasterio(dataset, area_id, name, proj_id=None, proj_dict=None):
    from pyresample.geometry import AreaDefinition

    a, b, c, d, e, f, _, _, _ = dataset.transform
    if not (b == d == 0):
        raise ValueError('Rotated rasters are not supported at this time.')

    if proj_dict is None:
        crs = dataset.crs
        if crs is not None:
            proj_dict = dataset.crs.to_dict()
        else:
            raise ValueError('The source raster is not gereferenced, please provide the value of proj_dict')

        if proj_id is None:
            proj_id = crs.wkt.split('"')[1]

    area_def = AreaDefinition(area_id, name, proj_id, proj_dict,
                              dataset.width, dataset.height, dataset.bounds)
    return area_def


def get_area_def_from_raster(source, area_id=None, name=None, proj_id=None, proj_dict=None):
    """Construct AreaDefinition object from raster.

    Parameters
    ----------
    source : str, Dataset, DatasetReader or DatasetWriter
        A file name. Also it can be ``osgeo.gdal.Dataset``,
        ``rasterio.io.DatasetReader`` or ``rasterio.io.DatasetWriter``
    area_id : str, optional
        ID of area
    name : str, optional
        Name of area
    proj_id : str, optional
        ID of projection
    proj_dict : dict, optional
        PROJ.4 parameters

    Returns
    -------
    area_def : object
        AreaDefinition object
    """
    try:
        import rasterio
    except ImportError:
        rasterio = None
        try:
            from osgeo import gdal
        except ImportError:
            raise ImportError('Either rasterio or gdal must be available')

    cleanup_gdal = cleanup_rasterio = None
    if isinstance(source, str):
        if rasterio is not None:
            source = rasterio.open(source)
            cleanup_rasterio = True
        else:
            source = gdal.Open(source)
            cleanup_gdal = True

    try:
        if rasterio is not None and isinstance(source, (rasterio.io.DatasetReader, rasterio.io.DatasetWriter)):
            return _get_area_def_from_rasterio(source, area_id, name, proj_id, proj_dict)
        return _get_area_def_from_gdal(source, area_id, name, proj_id, proj_dict)
    finally:
        if cleanup_rasterio:
            source.close()
        elif cleanup_gdal:
            source = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test kd_tree operations."""
import os
import unittest
from unittest import mock

import numpy as np

from pyresample import geometry, kd_tree, utils
from pyresample.test.utils import catch_warnings


class Test(unittest.TestCase):
    """Test nearest neighbor resampling on numpy arrays."""

    @classmethod
    def setUpClass(cls):
        cls.area_def = geometry.AreaDefinition('areaD',
                                               'Europe (3km, HRV, VTC)',
                                               'areaD',
                                               {'a': '6378144.0',
                                                'b': '6356759.0',
                                                'lat_0': '50.00',
                                                'lat_ts': '50.00',
                                                'lon_0': '8.00',
                                                'proj': 'stere'},
                                               800,
                                               800,
                                               [-1370912.72,
                                                   -909968.64000000001,
                                                   1029087.28,
                                                   1490031.3600000001])

        cls.tdata = np.array([1, 2, 3])
        cls.tlons = np.array([11.280789, 12.649354, 12.080402])
        cls.tlats = np.array([56.011037, 55.629675, 55.641535])
        cls.tswath = geometry.SwathDefinition(lons=cls.tlons, lats=cls.tlats)
        cls.tgrid = geometry.CoordinateDefinition(
            lons=np.array([12.562036]), lats=np.array([55.715613]))

    def test_nearest_base(self):
        res = kd_tree.resample_nearest(self.tswath,
                                       self.tdata.ravel(), self.tgrid,
                                       100000, reduce_data=False, segments=1)
        self.assertTrue(res[0] == 2)

    def test_gauss_base(self):
        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_gauss(self.tswath,
                                         self.tdata.ravel(), self.tgrid,
                                         50000, 25000, reduce_data=False, segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse(('Searching' not in str(w[0].message)))
        self.assertAlmostEqual(res[0], 2.2020729, 5)

    def test_custom_base(self):
        def wf(dist):
            return 1 - dist / 100000.0

        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_custom(self.tswath,
                                          self.tdata.ravel(), self.tgrid,
                                          50000, wf, reduce_data=False, segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse(('Searching' not in str(w[0].message)))
        self.assertAlmostEqual(res[0], 2.4356757, 5)

    def test_gauss_uncert(self):
        sigma = utils.fwhm2sigma(41627.730557884883)
        with catch_warnings(UserWarning) as w:
            res, stddev, count = kd_tree.resample_gauss(self.tswath, self.tdata,
                                                        self.tgrid, 100000, sigma,
                                                        with_uncert=True)
            self.assertTrue(len(w) > 0)
            self.assertTrue((any('Searching' in str(_w.message) for _w in w)))

        expected_res = 2.20206560694
        expected_stddev = 0.707115076173
        expected_count = 3
        self.assertAlmostEqual(res[0], expected_res, 5)
        self.assertAlmostEqual(stddev[0], expected_stddev, 5)
        self.assertEqual(count[0], expected_count)

    def test_custom_uncert(self):
        def wf(dist):
            return 1 - dist / 100000.0

        with catch_warnings(UserWarning) as w:
            res, stddev, counts = kd_tree.resample_custom(self.tswath,
                                                          self.tdata, self.tgrid,
                                                          100000, wf, with_uncert=True)
            self.assertTrue(len(w) > 0)
            self.assertTrue((any('Searching' in str(_w.message) for _w in w)))

        self.assertAlmostEqual(res[0], 2.32193149, 5)
        self.assertAlmostEqual(stddev[0], 0.81817972, 5)
        self.assertEqual(counts[0], 3)

    def test_nearest(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       self.area_def, 50000, segments=1)
        cross_sum = res.sum()
        expected = 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_masked_swath_target(self):
        """Test that a masked array works as a target."""
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        mask = np.ones_like(lons, dtype=np.bool_)
        mask[::2, ::2] = False
        swath_def = geometry.SwathDefinition(
            lons=np.ma.masked_array(lons, mask=mask),
            lats=np.ma.masked_array(lats, mask=False)
        )
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       swath_def, 50000, segments=3)
        cross_sum = res.sum()
        # expected = 12716  # if masks aren't respected
        expected = 12000
        self.assertEqual(cross_sum, expected)

    def test_nearest_1d(self):
        data = np.fromfunction(lambda x, y: x * y, (800, 800))
        lons = np.fromfunction(lambda x: 3 + x / 100., (500,))
        lats = np.fromfunction(lambda x: 75 - x / 10., (500,))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(self.area_def, data.ravel(),
                                       swath_def, 50000, segments=1)
        cross_sum = res.sum()
        expected = 35821299.0
        self.assertEqual(res.shape, (500,))
        self.assertEqual(cross_sum, expected)

    def test_nearest_empty(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 165 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       self.area_def, 50000, segments=1)
        cross_sum = res.sum()
        expected = 0
        self.assertEqual(cross_sum, expected)

    def test_nearest_empty_multi(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 165 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data_multi,
                                       self.area_def, 50000, segments=1)
        self.assertEqual(res.shape, (800, 800, 3),
                         msg='Swath resampling nearest empty multi failed')

    def test_nearest_empty_multi_masked(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 165 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data_multi,
                                       self.area_def, 50000, segments=1,
                                       fill_value=None)
        self.assertEqual(res.shape, (800, 800, 3))

    def test_nearest_empty_masked(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 165 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       self.area_def, 50000, segments=1,
                                       fill_value=None)
        cross_sum = res.mask.sum()
        expected = res.size
        self.assertTrue(cross_sum == expected)

    def test_nearest_segments(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       self.area_def, 50000, segments=2)
        cross_sum = res.sum()
        expected = 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_remap(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       self.area_def, 50000, segments=1)
        remap = kd_tree.resample_nearest(self.area_def, res.ravel(),
                                         swath_def, 5000, segments=1)
        cross_sum = remap.sum()
        expected = 22275.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_mp(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       self.area_def, 50000, nprocs=2, segments=1)
        cross_sum = res.sum()
        expected = 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_multi(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        res = kd_tree.resample_nearest(swath_def, data_multi,
                                       self.area_def, 50000, segments=1)
        cross_sum = res.sum()
        expected = 3 * 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_multi_unraveled(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.dstack((data, data, data))
        res = kd_tree.resample_nearest(swath_def, data_multi,
                                       self.area_def, 50000, segments=1)
        cross_sum = res.sum()
        expected = 3 * 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_gauss_sparse(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_gauss(swath_def, data.ravel(),
                                     self.area_def, 50000, 25000, fill_value=-1, segments=1)
        cross_sum = res.sum()
        expected = 15387753.9852
        self.assertAlmostEqual(cross_sum, expected, places=3)

    def test_gauss(self):
        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -5, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_gauss(swath_def, data.ravel(),
                                         self.area_def, 50000, 25000, segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse(('Possible more' not in str(w[0].message)))
        cross_sum = res.sum()
        expected = 4872.8100353517921
        self.assertAlmostEqual(cross_sum, expected)

    def test_gauss_fwhm(self):
        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -5, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_gauss(swath_def, data.ravel(),
                                         self.area_def, 50000, utils.fwhm2sigma(41627.730557884883), segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse(('Possible more' not in str(w[0].message)))
        cross_sum = res.sum()
        expected = 4872.8100353517921
        self.assertAlmostEqual(cross_sum, expected)

    def test_gauss_multi(self):
        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -6, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_gauss(swath_def, data_multi,
                                         self.area_def, 50000, [25000, 15000, 10000], segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse(('Possible more' not in str(w[0].message)))
        cross_sum = res.sum()
        expected = 1461.8429990248171
        self.assertAlmostEqual(cross_sum, expected)

    def test_gauss_multi_uncert(self):
        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -6, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        with catch_warnings(UserWarning) as w:
            # The assertion below checks if there is only one warning raised
            # and whether it contains a specific message from pyresample
            # On python 2.7.9+ the resample_gauss method raises multiple deprecation warnings
            # that cause to fail, so we ignore the unrelated warnings.
            res, stddev, counts = kd_tree.resample_gauss(swath_def, data_multi,
                                                         self.area_def, 50000, [
                                                             25000, 15000, 10000],
                                                         segments=1, with_uncert=True)
            self.assertTrue(len(w) >= 1)
            self.assertTrue(
                any(['Possible more' in str(x.message) for x in w]))
        cross_sum = res.sum()
        cross_sum_counts = counts.sum()
        expected = 1461.8429990248171
        expected_stddev = [0.44621800779801657, 0.44363137712896705,
                           0.43861019464274459]
        expected_counts = 4934802.0
        self.assertTrue(res.shape == stddev.shape and stddev.shape == counts.shape and counts.shape == (800, 800, 3))
        self.assertAlmostEqual(cross_sum, expected)

        for i, e_stddev in enumerate(expected_stddev):
            cross_sum_stddev = stddev[:, :, i].sum()
            self.assertAlmostEqual(cross_sum_stddev, e_stddev)
        self.assertAlmostEqual(cross_sum_counts, expected_counts)

    def test_gauss_multi_mp(self):
        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -6, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_gauss(swath_def, data_multi,
                                         self.area_def, 50000, [
                                             25000, 15000, 10000],
                                         nprocs=2, segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse(('Possible more' not in str(w[0].message)))
        cross_sum = res.sum()
        expected = 1461.8429990248171
        self.assertAlmostEqual(cross_sum, expected)

    def test_gauss_multi_mp_segments(self):
        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -6, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_gauss(swath_def, data_multi,
                                         self.area_def, 50000, [
                                             25000, 15000, 10000],
                                         nprocs=2, segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse('Possible more' not in str(w[0].message))
        cross_sum = res.sum()
        expected = 1461.8429990248171
        self.assertAlmostEqual(cross_sum, expected)

    def test_gauss_multi_mp_segments_empty(self):
        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -6, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 165 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        res = kd_tree.resample_gauss(swath_def, data_multi,
                                     self.area_def, 50000, [
                                         25000, 15000, 10000],
                                     nprocs=2, segments=1)
        cross_sum = res.sum()
        self.assertTrue(cross_sum == 0)

    def test_custom(self):
        def wf(dist):
            return 1 - dist / 100000.0

        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -5, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_custom(swath_def, data.ravel(),
                                          self.area_def, 50000, wf, segments=1)
            # PyProj proj/CRS and "more than 8 neighbours" are warned about
            self.assertFalse(len(w) > 2)
            neighbour_warn = False
            for warn in w:
                if 'Possible more' in str(warn.message):
                    neighbour_warn = True
                    break
            self.assertTrue(neighbour_warn)
            if len(w) == 2:
                proj_crs_warn = False
                for warn in w:
                    if 'important projection information' in str(warn.message):
                        proj_crs_warn = True
                        break
                self.assertTrue(proj_crs_warn)

        cross_sum = res.sum()
        expected = 4872.8100347930776
        self.assertAlmostEqual(cross_sum, expected)

    def test_custom_multi(self):
        def wf1(dist):
            return 1 - dist / 100000.0

        def wf2(dist):
            return 1

        def wf3(dist):
            return np.cos(dist) ** 2

        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -6, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))
        with catch_warnings(UserWarning) as w:
            res = kd_tree.resample_custom(swath_def, data_multi,
                                          self.area_def, 50000, [wf1, wf2, wf3], segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse('Possible more' not in str(w[0].message))
        cross_sum = res.sum()
        expected = 1461.8428378742638
        self.assertAlmostEqual(cross_sum, expected)

    def test_masked_nearest(self):
        data = np.ones((50, 10))
        data[:, 5:] = 2
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        mask = np.ones((50, 10))
        mask[:, :5] = 0
        masked_data = np.ma.array(data, mask=mask)
        res = kd_tree.resample_nearest(swath_def, masked_data.ravel(),
                                       self.area_def, 50000, segments=1)
        expected_mask = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                 'test_files',
                                                 'mask_test_nearest_mask.dat'),
                                    sep=' ').reshape((800, 800))
        expected_data = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                 'test_files',
                                                 'mask_test_nearest_data.dat'),
                                    sep=' ').reshape((800, 800))
        self.assertTrue(np.array_equal(expected_mask, res.mask))
        self.assertTrue(np.array_equal(expected_data, res.data))

    def test_masked_nearest_1d(self):
        data = np.ones((800, 800))
        data[:400, :] = 2
        lons = np.fromfunction(lambda x: 3 + x / 100., (500,))
        lats = np.fromfunction(lambda x: 75 - x / 10., (500,))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        mask = np.ones((800, 800))
        mask[400:, :] = 0
        masked_data = np.ma.array(data, mask=mask)
        res = kd_tree.resample_nearest(self.area_def, masked_data.ravel(),
                                       swath_def, 50000, segments=1)
        self.assertEqual(res.mask.sum(), 112)

    def test_masked_gauss(self):
        data = np.ones((50, 10))
        data[:, 5:] = 2
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        mask = np.ones((50, 10))
        mask[:, :5] = 0
        masked_data = np.ma.array(data, mask=mask)
        res = kd_tree.resample_gauss(swath_def, masked_data.ravel(),
                                     self.area_def, 50000, 25000, segments=1)
        expected_mask = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                 'test_files',
                                                 'mask_test_mask.dat'),
                                    sep=' ').reshape((800, 800))
        expected_data = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                 'test_files',
                                                 'mask_test_data.dat'),
                                    sep=' ').reshape((800, 800))
        expected = expected_data.sum()
        cross_sum = res.data.sum()

        self.assertTrue(np.array_equal(expected_mask, res.mask))
        self.assertAlmostEqual(cross_sum, expected, places=3)

    def test_masked_fill_float(self):
        data = np.ones((50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       self.area_def, 50000, fill_value=None, segments=1)
        expected_fill_mask = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                      'test_files',
                                                      'mask_test_fill_value.dat'),
                                         sep=' ').reshape((800, 800))
        fill_mask = res.mask
        self.assertTrue(np.array_equal(fill_mask, expected_fill_mask))

    def test_masked_fill_int(self):
        data = np.ones((50, 10)).astype('int')
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def, data.ravel(),
                                       self.area_def, 50000, fill_value=None, segments=1)
        expected_fill_mask = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                      'test_files',
                                                      'mask_test_fill_value.dat'),
                                         sep=' ').reshape((800, 800))
        fill_mask = res.mask
        self.assertTrue(np.array_equal(fill_mask, expected_fill_mask))

    def test_masked_full(self):
        data = np.ones((50, 10))
        data[:, 5:] = 2
        mask = np.ones((50, 10))
        mask[:, :5] = 0
        masked_data = np.ma.array(data, mask=mask)
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def,
                                       masked_data.ravel(
                                       ), self.area_def, 50000,
                                       fill_value=None, segments=1)
        expected_fill_mask = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                      'test_files',
                                                      'mask_test_full_fill.dat'),
                                         sep=' ').reshape((800, 800))
        fill_mask = res.mask

        self.assertTrue(np.array_equal(fill_mask, expected_fill_mask))

    def test_masked_full_multi(self):
        data = np.ones((50, 10))
        data[:, 5:] = 2
        mask1 = np.ones((50, 10))
        mask1[:, :5] = 0
        mask2 = np.ones((50, 10))
        mask2[:, 5:] = 0
        mask3 = np.ones((50, 10))
        mask3[:25, :] = 0
        data_multi = np.column_stack(
            (data.ravel(), data.ravel(), data.ravel()))
        mask_multi = np.column_stack(
            (mask1.ravel(), mask2.ravel(), mask3.ravel()))
        masked_data = np.ma.array(data_multi, mask=mask_multi)
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        res = kd_tree.resample_nearest(swath_def,
                                       masked_data, self.area_def, 50000,
                                       fill_value=None, segments=1)
        expected_fill_mask = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                      'test_files',
                                                      'mask_test_full_fill_multi.dat'),
                                         sep=' ').reshape((800, 800, 3))
        fill_mask = res.mask
        cross_sum = res.sum()
        expected = 357140.0
        self.assertAlmostEqual(cross_sum, expected)
        self.assertTrue(np.array_equal(fill_mask, expected_fill_mask))

    def test_dtype(self):
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        grid_def = geometry.GridDefinition(lons, lats)
        lons = np.asarray(lons, dtype='f4')
        lats = np.asarray(lats, dtype='f4')
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        valid_input_index, valid_output_index, index_array, distance_array = \
            kd_tree.get_neighbour_info(swath_def,
                                       grid_def,
                                       50000, neighbours=1, segments=1)

    def test_nearest_from_sample(self):
        data = np.fromfunction(lambda y, x: y * x, (50, 10))
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        valid_input_index, valid_output_index, index_array, distance_array = \
            kd_tree.get_neighbour_info(swath_def,
                                       self.area_def,
                                       50000, neighbours=1, segments=1)
        res = kd_tree.get_sample_from_neighbour_info('nn', (800, 800), data.ravel(),
                                                     valid_input_index, valid_output_index,
                                                     index_array)
        cross_sum = res.sum()
        expected = 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_from_sample_np_dtypes(self):
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        valid_input_index, valid_output_index, index_array, distance_array = \
            kd_tree.get_neighbour_info(swath_def,
                                       self.area_def,
                                       50000, neighbours=1, segments=1)

        for dtype in [np.uint16, np.float32]:
            with self.subTest(dtype):
                data = np.fromfunction(lambda y, x: y * x, (50, 10)).astype(dtype)
                fill_value = dtype(0.0)
                res = \
                    kd_tree.get_sample_from_neighbour_info('nn', (800, 800),
                                                           data.ravel(),
                                                           valid_input_index,
                                                           valid_output_index,
                                                           index_array,
                                                           fill_value=fill_value)
                cross_sum = res.sum()
                expected = 15874591.0
                self.assertEqual(cross_sum, expected)

    def test_custom_multi_from_sample(self):
        def wf1(dist):
            return 1 - dist / 100000.0

        def wf2(dist):
            return 1

        def wf3(dist):
            return np.cos(dist) ** 2

        data = np.fromfunction(lambda y, x: (y + x) * 10 ** -6, (5000, 100))
        lons = np.fromfunction(
            lambda y, x: 3 + (10.0 / 100) * x, (5000, 100))
        lats = np.fromfunction(
            lambda y, x: 75 - (50.0 / 5000) * y, (5000, 100))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        data_multi = np.column_stack((data.ravel(), data.ravel(),
                                      data.ravel()))

        with catch_warnings(UserWarning) as w:
            valid_input_index, valid_output_index, index_array, distance_array = \
                kd_tree.get_neighbour_info(swath_def,
                                           self.area_def,
                                           50000, segments=1)
            self.assertFalse(len(w) != 1)
            self.assertFalse(('Possible more' not in str(w[0].message)))

        res = kd_tree.get_sample_from_neighbour_info('custom', (800, 800),
                                                     data_multi,
                                                     valid_input_index, valid_output_index,
                                                     index_array, distance_array,
                                                     weight_funcs=[wf1, wf2, wf3])

        cross_sum = res.sum()

        expected = 1461.8428378742638
        self.assertAlmostEqual(cross_sum, expected)
        res = kd_tree.get_sample_from_neighbour_info('custom', (800, 800),
                                                     data_multi,
                                                     valid_input_index, valid_output_index,
                                                     index_array, distance_array,
                                                     weight_funcs=[wf1, wf2, wf3])

        # Look for error where input data has been manipulated
        cross_sum = res.sum()
        expected = 1461.8428378742638
        self.assertAlmostEqual(cross_sum, expected)

    def test_masked_multi_from_sample(self):
        data = np.ones((50, 10))
        data[:, 5:] = 2
        mask1 = np.ones((50, 10))
        mask1[:, :5] = 0
        mask2 = np.ones((50, 10))
        mask2[:, 5:] = 0
        mask3 = np.ones((50, 10))
        mask3[:25, :] = 0
        data_multi = np.column_stack(
            (data.ravel(), data.ravel(), data.ravel()))
        mask_multi = np.column_stack(
            (mask1.ravel(), mask2.ravel(), mask3.ravel()))
        masked_data = np.ma.array(data_multi, mask=mask_multi)
        lons = np.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = np.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        valid_input_index, valid_output_index, index_array, distance_array = \
            kd_tree.get_neighbour_info(swath_def,
                                       self.area_def,
                                       50000, neighbours=1, segments=1)
        res = kd_tree.get_sample_from_neighbour_info('nn', (800, 800),
                                                     masked_data,
                                                     valid_input_index,
                                                     valid_output_index, index_array,
                                                     fill_value=None)
        expected_fill_mask = np.fromfile(os.path.join(os.path.dirname(__file__),
                                                      'test_files',
                                                      'mask_test_full_fill_multi.dat'),
                                         sep=' ').reshape((800, 800, 3))
        fill_mask = res.mask
        self.assertTrue(np.array_equal(fill_mask, expected_fill_mask))


class TestXArrayResamplerNN(unittest.TestCase):
    """Test the XArrayResamplerNN class."""

    @classmethod
    def setUpClass(cls):
        import dask.array as da
        import xarray as xr
        cls.area_def = geometry.AreaDefinition('areaD',
                                               'Europe (3km, HRV, VTC)',
                                               'areaD',
                                               {'a': '6378144.0',
                                                'b': '6356759.0',
                                                'lat_0': '50.00',
                                                'lat_ts': '50.00',
                                                'lon_0': '8.00',
                                                'proj': 'stere'},
                                               800,
                                               800,
                                               [-1370912.72,
                                                -909968.64000000001,
                                                1029087.28,
                                                1490031.3600000001])

        dfa = da.from_array  # shortcut
        cls.chunks = chunks = 5
        cls.tgrid = geometry.CoordinateDefinition(
            lons=dfa(np.array([
                [11.5, 12.562036, 12.9],
                [11.5, 12.562036, 12.9],
                [11.5, 12.562036, 12.9],
                [11.5, 12.562036, 12.9],
            ]), chunks=chunks),
            lats=dfa(np.array([
                [55.715613, 55.715613, 55.715613],
                [55.715613, 55.715613, 55.715613],
                [55.715613, np.nan, 55.715613],
                [55.715613, 55.715613, 55.715613],
            ]), chunks=chunks))

        cls.tdata_1d = xr.DataArray(
            dfa(np.array([1., 2., 3.]), chunks=chunks), dims=('my_dim1',))
        cls.tlons_1d = xr.DataArray(
            dfa(np.array([11.280789, 12.649354, 12.080402]), chunks=chunks),
            dims=('my_dim1',))
        cls.tlats_1d = xr.DataArray(
            dfa(np.array([56.011037, 55.629675, 55.641535]), chunks=chunks),
            dims=('my_dim1',))
        cls.tswath_1d = geometry.SwathDefinition(lons=cls.tlons_1d,
                                                 lats=cls.tlats_1d)

        cls.data_2d = xr.DataArray(
            da.from_array(np.fromfunction(lambda y, x: y * x, (50, 10)),
                          chunks=5),
            dims=('my_dim_y', 'my_dim_x'))
        cls.data_3d = xr.DataArray(
            da.from_array(np.fromfunction(lambda y, x, b: y * x * b, (50, 10, 3)),
                          chunks=5),
            dims=('my_dim_y', 'my_dim_x', 'bands'),
            coords={'bands': ['r', 'g', 'b']})
        cls.lons_2d = xr.DataArray(
            da.from_array(np.fromfunction(lambda y, x: 3 + x, (50, 10)),
                          chunks=5),
            dims=('my_dim_y', 'my_dim_x'))
        cls.lats_2d = xr.DataArray(
            da.from_array(np.fromfunction(lambda y, x: 75 - y, (50, 10)),
                          chunks=5),
            dims=('my_dim_y', 'my_dim_x'))
        cls.swath_def_2d = geometry.SwathDefinition(lons=cls.lons_2d,
                                                    lats=cls.lats_2d)
        cls.src_area_2d = geometry.AreaDefinition(
            'areaD_src', 'Europe (3km, HRV, VTC)', 'areaD',
            {'a': '6378144.0', 'b': '6356759.0', 'lat_0': '52.00',
             'lat_ts': '52.00', 'lon_0': '5.00', 'proj': 'stere'}, 50, 10,
            [-1370912.72, -909968.64000000001, 1029087.28,
             1490031.3600000001])

    def test_nearest_swath_1d_mask_to_grid_1n(self):
        """Test 1D swath definition to 2D grid definition; 1 neighbor."""
        import dask.array as da
        import xarray as xr

        from pyresample.kd_tree import XArrayResamplerNN
        resampler = XArrayResamplerNN(self.tswath_1d, self.tgrid,
                                      radius_of_influence=100000,
                                      neighbours=1)
        data = self.tdata_1d
        ninfo = resampler.get_neighbour_info(mask=data.isnull())
        for val in ninfo[:3]:
            # vii, ia, voi
            self.assertIsInstance(val, da.Array)
        res = resampler.get_sample_from_neighbour_info(data)
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        actual = res.values
        expected = np.array([
            [1., 2., 2.],
            [1., 2., 2.],
            [1., np.nan, 2.],
            [1., 2., 2.],
        ])
        np.testing.assert_allclose(actual, expected)

    def test_nearest_type_preserve(self):
        """Test 1D swath definition to 2D grid definition; 1 neighbor."""
        import dask.array as da
        import xarray as xr

        from pyresample.kd_tree import XArrayResamplerNN
        resampler = XArrayResamplerNN(self.tswath_1d, self.tgrid,
                                      radius_of_influence=100000,
                                      neighbours=1)
        data = self.tdata_1d
        data = xr.DataArray(da.from_array(np.array([1, 2, 3]),
                                          chunks=5),
                            dims=('my_dim1',))
        ninfo = resampler.get_neighbour_info()
        for val in ninfo[:3]:
            # vii, ia, voi
            self.assertIsInstance(val, da.Array)
        res = resampler.get_sample_from_neighbour_info(data, fill_value=255)
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        actual = res.values
        expected = np.array([
            [1, 2, 2],
            [1, 2, 2],
            [1, 255, 2],
            [1, 2, 2],
        ])
        np.testing.assert_equal(actual, expected)

    def test_nearest_swath_2d_mask_to_area_1n(self):
        """Test 2D swath definition to 2D area definition; 1 neighbor."""
        import dask.array as da
        import xarray as xr

        from pyresample.kd_tree import XArrayResamplerNN
        swath_def = self.swath_def_2d
        data = self.data_2d
        resampler = XArrayResamplerNN(swath_def, self.area_def,
                                      radius_of_influence=50000,
                                      neighbours=1)
        ninfo = resampler.get_neighbour_info(mask=data.isnull())
        for val in ninfo[:3]:
            # vii, ia, voi
            self.assertIsInstance(val, da.Array)
        res = resampler.get_sample_from_neighbour_info(data)
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        res = res.values
        cross_sum = np.nansum(res)
        expected = 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_area_2d_to_area_1n(self):
        """Test 2D area definition to 2D area definition; 1 neighbor."""
        import dask.array as da
        import xarray as xr

        from pyresample.kd_tree import XArrayResamplerNN
        from pyresample.test.utils import assert_maximum_dask_computes
        data = self.data_2d
        resampler = XArrayResamplerNN(self.src_area_2d, self.area_def,
                                      radius_of_influence=50000,
                                      neighbours=1)
        with assert_maximum_dask_computes(0):
            ninfo = resampler.get_neighbour_info()
        for val in ninfo[:3]:
            # vii, ia, voi
            self.assertIsInstance(val, da.Array)
        self.assertRaises(AssertionError,
                          resampler.get_sample_from_neighbour_info, data)

        # rename data dimensions to match the expected area dimensions
        data = data.rename({'my_dim_y': 'y', 'my_dim_x': 'x'})
        with assert_maximum_dask_computes(0):
            res = resampler.get_sample_from_neighbour_info(data)
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        res = res.values
        cross_sum = np.nansum(res)
        expected = 27706753.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_area_2d_to_area_1n_no_roi(self):
        """Test 2D area definition to 2D area definition; 1 neighbor, no radius of influence."""
        import dask.array as da
        import xarray as xr

        from pyresample.kd_tree import XArrayResamplerNN
        data = self.data_2d
        resampler = XArrayResamplerNN(self.src_area_2d, self.area_def,
                                      neighbours=1)
        ninfo = resampler.get_neighbour_info()
        for val in ninfo[:3]:
            # vii, ia, voi
            self.assertIsInstance(val, da.Array)
        self.assertRaises(AssertionError,
                          resampler.get_sample_from_neighbour_info, data)

        # rename data dimensions to match the expected area dimensions
        data = data.rename({'my_dim_y': 'y', 'my_dim_x': 'x'})
        res = resampler.get_sample_from_neighbour_info(data)
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        res = res.values
        cross_sum = np.nansum(res)
        expected = 87281406.0
        self.assertEqual(cross_sum, expected)

        # pretend the resolutions can't be determined
        with mock.patch.object(self.src_area_2d, 'geocentric_resolution') as sgr, \
                mock.patch.object(self.area_def, 'geocentric_resolution') as dgr:
            sgr.side_effect = RuntimeError
            dgr.side_effect = RuntimeError
            resampler = XArrayResamplerNN(self.src_area_2d, self.area_def,
                                          neighbours=1)
            resampler.get_neighbour_info()
            res = resampler.get_sample_from_neighbour_info(data)
            self.assertIsInstance(res, xr.DataArray)
            self.assertIsInstance(res.data, da.Array)
            res = res.values
            cross_sum = np.nansum(res)
            expected = 1855928.0
            self.assertEqual(cross_sum, expected)

    def test_nearest_area_2d_to_area_1n_3d_data(self):
        """Test 2D area definition to 2D area definition; 1 neighbor, 3d data."""
        import dask.array as da
        import xarray as xr

        from pyresample.kd_tree import XArrayResamplerNN
        data = self.data_3d
        resampler = XArrayResamplerNN(self.src_area_2d, self.area_def,
                                      radius_of_influence=50000,
                                      neighbours=1)
        ninfo = resampler.get_neighbour_info()
        for val in ninfo[:3]:
            # vii, ia, voi
            self.assertIsInstance(val, da.Array)
        self.assertRaises(AssertionError,
                          resampler.get_sample_from_neighbour_info, data)

        # rename data dimensions to match the expected area dimensions
        data = data.rename({'my_dim_y': 'y', 'my_dim_x': 'x'})
        res = resampler.get_sample_from_neighbour_info(data)
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        self.assertCountEqual(res.coords['bands'], ['r', 'g', 'b'])
        res = res.values
        cross_sum = np.nansum(res)
        expected = 83120259.0
        self.assertEqual(cross_sum, expected)

    @unittest.skipIf(True, "Multiple neighbors not supported yet")
    def test_nearest_swath_1d_mask_to_grid_8n(self):
        """Test 1D swath definition to 2D grid definition; 8 neighbors."""
        import dask.array as da
        import xarray as xr

        from pyresample.kd_tree import XArrayResamplerNN
        resampler = XArrayResamplerNN(self.tswath_1d, self.tgrid,
                                      radius_of_influence=100000,
                                      neighbours=8)
        data = self.tdata_1d
        ninfo = resampler.get_neighbour_info(mask=data.isnull())
        for val in ninfo[:3]:
            # vii, ia, voi
            self.assertIsInstance(val, da.Array)
        res = resampler.get_sample_from_neighbour_info(data)
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        # actual = res.values
        # expected = TODO
        # np.testing.assert_allclose(actual, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2021 Pyresample Developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test grid interface."""
import unittest

import numpy as np

from pyresample import geometry, grid, utils


class Test(unittest.TestCase):
    """Test grid interface."""

    area_def = geometry.AreaDefinition('areaD', 'Europe (3km, HRV, VTC)', 'areaD',
                                       {'a': '6378144.0',
                                        'b': '6356759.0',
                                        'lat_0': '50.00',
                                        'lat_ts': '50.00',
                                        'lon_0': '8.00',
                                        'proj': 'stere'},
                                       800,
                                       800,
                                       [-1370912.72,
                                           -909968.64000000001,
                                           1029087.28,
                                           1490031.3600000001])

    area_def2 = geometry.AreaDefinition('areaD2', 'Europe (3km, HRV, VTC)', 'areaD2',
                                        {'a': '6378144.0',
                                         'b': '6356759.0',
                                         'lat_0': '50.00',
                                         'lat_ts': '50.00',
                                         'lon_0': '8.00',
                                         'proj': 'stere'},
                                        5,
                                        5,
                                        [-1370912.72,
                                            -909968.64000000001,
                                            1029087.28,
                                            1490031.3600000001])

    msg_area = geometry.AreaDefinition('msg_full', 'Full globe MSG image 0 degrees',
                                       'msg_full',
                                       {'a': '6378169.0',
                                        'b': '6356584.0',
                                        'h': '35785831.0',
                                        'lon_0': '0',
                                        'proj': 'geos'},
                                       3712,
                                       3712,
                                       [-5568742.4000000004,
                                           -5568742.4000000004,
                                           5568742.4000000004,
                                           5568742.4000000004]
                                       )

    def test_linesample(self):
        data = np.fromfunction(lambda y, x: y * x, (40, 40))
        rows = np.array([[1, 2], [3, 4]])
        cols = np.array([[25, 26], [27, 28]])
        res = grid.get_image_from_linesample(rows, cols, data)
        expected = np.array([[25., 52.], [81., 112.]])
        self.assertTrue(np.array_equal(res, expected), 'Linesample failed')

    def test_linesample_multi(self):
        data1 = np.fromfunction(lambda y, x: y * x, (40, 40))
        data2 = np.fromfunction(lambda y, x: 2 * y * x, (40, 40))
        data3 = np.fromfunction(lambda y, x: 3 * y * x, (40, 40))
        data = np.zeros((40, 40, 3))
        data[:, :, 0] = data1
        data[:, :, 1] = data2
        data[:, :, 2] = data3
        rows = np.array([[1, 2], [3, 4]])
        cols = np.array([[25, 26], [27, 28]])
        res = grid.get_image_from_linesample(rows, cols, data)
        expected = np.array([[[25., 50., 75.],
                              [52., 104., 156.]],
                             [[81., 162., 243.],
                              [112., 224., 336.]]])
        self.assertTrue(np.array_equal(res, expected), 'Linesample failed')

    def test_from_latlon(self):
        data = np.fromfunction(lambda y, x: y * x, (800, 800))
        lons = np.fromfunction(lambda y, x: x, (10, 10))
        lats = np.fromfunction(lambda y, x: 50 - (5.0 / 10) * y, (10, 10))
        source_def = self.area_def
        res = grid.get_image_from_lonlats(lons, lats, source_def, data)
        expected = np.array([[129276., 141032., 153370., 165804., 178334., 190575.,
                              202864., 214768., 226176., 238080.],
                             [133056., 146016., 158808., 171696., 184320., 196992.,
                              209712., 222480., 234840., 247715.],
                             [137026., 150150., 163370., 177215., 190629., 203756.,
                              217464., 230256., 243048., 256373.],
                             [140660., 154496., 168714., 182484., 196542., 210650.,
                              224257., 238464., 251712., 265512.],
                             [144480., 158484., 173148., 187912., 202776., 217358.,
                              231990., 246240., 259920., 274170.],
                             [147968., 163261., 178398., 193635., 208616., 223647.,
                              238728., 253859., 268584., 283898.],
                             [151638., 167121., 182704., 198990., 214775., 230280.,
                              246442., 261617., 276792., 292574.],
                             [154980., 171186., 187860., 204016., 220542., 237120.,
                              253125., 269806., 285456., 301732.],
                             [158500., 175536., 192038., 209280., 226626., 243697.,
                              260820., 277564., 293664., 310408.],
                             [161696., 179470., 197100., 214834., 232320., 250236.,
                              267448., 285090., 302328., 320229.]])
        self.assertTrue(
            np.array_equal(res, expected), 'Sampling from lat lon failed')

    def test_proj_coords(self):
        res = self.area_def2.get_proj_coords()
        cross_sum = res[0].sum() + res[1].sum()
        expected = 2977965.9999999963
        self.assertAlmostEqual(
            cross_sum, expected, msg='Calculation of proj coords failed')

    def test_latlons(self):
        res = self.area_def2.get_lonlats()
        cross_sum = res[0].sum() + res[1].sum()
        expected = 1440.8280578215431
        self.assertAlmostEqual(
            cross_sum, expected, msg='Calculation of lat lons failed')

    def test_latlons_mp(self):
        res = self.area_def2.get_lonlats(nprocs=2)
        cross_sum = res[0].sum() + res[1].sum()
        expected = 1440.8280578215431
        self.assertAlmostEqual(
            cross_sum, expected, msg='Calculation of lat lons failed')

    def test_resampled_image(self):
        data = np.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        target_def = self.area_def
        source_def = self.msg_area
        res = grid.get_resampled_image(
            target_def, source_def, data, segments=1)
        cross_sum = res.sum()
        expected = 399936.39392500359
        self.assertAlmostEqual(
            cross_sum, expected, msg='Resampling of image failed')

    def test_resampled_image_masked(self):
        # Generate test image with masked elements
        data = np.ma.ones(self.msg_area.shape)
        data.mask = np.zeros(data.shape)
        data.mask[253:400, 1970:2211] = 1

        # Resample image using multiple segments
        target_def = self.area_def
        source_def = self.msg_area
        res = grid.get_resampled_image(
            target_def, source_def, data, segments=4, fill_value=None)

        # Make sure the mask has been preserved
        self.assertGreater(res.mask.sum(), 0,
                           msg='Resampling did not preserve the mask')

    def test_generate_linesample(self):
        data = np.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        row_indices, col_indices = utils.generate_quick_linesample_arrays(self.msg_area,
                                                                          self.area_def)
        res = data[row_indices, col_indices]
        cross_sum = res.sum()
        expected = 399936.39392500359
        self.assertAlmostEqual(
            cross_sum, expected, msg='Generate linesample failed')
        self.assertFalse(row_indices.dtype != np.uint16 or col_indices.dtype != np.uint16,
                         'Generate linesample failed. Downcast to uint16 expected')

    def test_resampled_image_mp(self):
        data = np.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        target_def = self.area_def
        source_def = self.msg_area
        res = grid.get_resampled_image(
            target_def, source_def, data, nprocs=2, segments=1)
        cross_sum = res.sum()
        expected = 399936.39392500359
        self.assertAlmostEqual(
            cross_sum, expected, msg='Resampling of image mp failed')

    def test_single_lonlat(self):
        lon, lat = self.area_def.get_lonlat(400, 400)
        self.assertAlmostEqual(
            lon, 5.5028467120975835, msg='Resampling of single lon failed')
        self.assertAlmostEqual(
            lat, 52.566998432390619, msg='Resampling of single lat failed')

    def test_proj4_string(self):
        """Test 'proj_str' property of AreaDefinition."""
        proj4_string = self.area_def.proj_str
        # different versions of PROJ/pyproj simplify parameters in
        # different ways. Just check for the minimum expected.
        param_strings = (
            "+a=6378144",
            "+lon_0=8",
            "+lat_0=50",
            "+proj=stere",
        )
        for param_str in param_strings:
            assert param_str in proj4_string



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2021 Pyresample Developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test image interface."""

import os
import unittest

import numpy

from pyresample import geometry, image, utils


class Test(unittest.TestCase):
    """Test image interface."""

    area_def = geometry.AreaDefinition('areaD', 'Europe (3km, HRV, VTC)',
                                       'areaD',
                                       {'a': '6378144.0',
                                        'b': '6356759.0',
                                        'lat_0': '50.00',
                                        'lat_ts': '50.00',
                                        'lon_0': '8.00',
                                        'proj': 'stere'},
                                       800,
                                       800,
                                       [-1370912.72,
                                        -909968.64000000001,
                                        1029087.28,
                                        1490031.3600000001])

    msg_area = geometry.AreaDefinition('msg_full',
                                       'Full globe MSG image 0 degrees',
                                       'msg_full',
                                       {'a': '6378169.0',
                                        'b': '6356584.0',
                                        'h': '35785831.0',
                                        'lon_0': '0',
                                        'proj': 'geos'},
                                       3712,
                                       3712,
                                       [-5568742.4000000004,
                                        -5568742.4000000004,
                                        5568742.4000000004,
                                        5568742.4000000004])

    msg_area_resize = geometry.AreaDefinition('msg_full',
                                              'Full globe MSG image 0 degrees',
                                              'msg_full',
                                              {'a': '6378169.0',
                                               'b': '6356584.0',
                                               'h': '35785831.0',
                                               'lon_0': '0',
                                               'proj': 'geos'},
                                              928,
                                              928,
                                              [-5568742.4000000004,
                                               -5568742.4000000004,
                                               5568742.4000000004,
                                               5568742.4000000004])

    def test_image(self):
        data = numpy.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        msg_con = image.ImageContainerQuick(data, self.msg_area, segments=1)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum = res.sum()
        expected = 399936.39392500359
        self.assertAlmostEqual(cross_sum, expected)

    def test_image_segments(self):
        data = numpy.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        msg_con = image.ImageContainerQuick(data, self.msg_area, segments=8)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum = res.sum()
        expected = 399936.39392500359
        self.assertAlmostEqual(cross_sum, expected)

    def test_return_type(self):
        data = numpy.ones((3712, 3712)).astype('int')
        msg_con = image.ImageContainerQuick(data, self.msg_area, segments=1)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        self.assertTrue(data.dtype is res.dtype)

    def test_masked_image(self):
        data = numpy.zeros((3712, 3712))
        mask = numpy.zeros((3712, 3712))
        mask[:, 1865:] = 1
        data_masked = numpy.ma.array(data, mask=mask)
        msg_con = image.ImageContainerQuick(
            data_masked, self.msg_area, segments=1)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        resampled_mask = res.mask.astype('int')
        expected = numpy.fromfile(os.path.join(os.path.dirname(__file__),
                                               'test_files', 'mask_grid.dat'),
                                  sep=' ').reshape((800, 800))
        self.assertTrue(numpy.array_equal(resampled_mask, expected))

    def test_masked_image_fill(self):
        data = numpy.zeros((3712, 3712))
        mask = numpy.zeros((3712, 3712))
        mask[:, 1865:] = 1
        data_masked = numpy.ma.array(data, mask=mask)
        msg_con = image.ImageContainerQuick(data_masked, self.msg_area,
                                            fill_value=None, segments=1)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        resampled_mask = res.mask.astype('int')
        expected = numpy.fromfile(os.path.join(os.path.dirname(__file__),
                                               'test_files',
                                               'mask_grid.dat'),
                                  sep=' ').reshape((800, 800))
        self.assertTrue(numpy.array_equal(resampled_mask, expected))

    def test_nearest_neighbour(self):
        data = numpy.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        msg_con = image.ImageContainerNearest(
            data, self.msg_area, 50000, segments=1)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum = res.sum()
        expected = 399936.70287099993
        self.assertAlmostEqual(cross_sum, expected)

    def test_nearest_resize(self):
        data = numpy.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        msg_con = image.ImageContainerNearest(
            data, self.msg_area, 50000, segments=1)
        area_con = msg_con.resample(self.msg_area_resize)
        res = area_con.image_data
        cross_sum = res.sum()
        expected = 2212023.0175830
        self.assertAlmostEqual(cross_sum, expected)

    def test_nearest_neighbour_multi(self):
        data1 = numpy.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        data2 = numpy.fromfunction(
            lambda y, x: y * x * 10 ** -6, (3712, 3712)) * 2
        data = numpy.dstack((data1, data2))
        msg_con = image.ImageContainerNearest(
            data, self.msg_area, 50000, segments=1)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum1 = res[:, :, 0].sum()
        expected1 = 399936.70287099993
        self.assertAlmostEqual(cross_sum1, expected1)

        cross_sum2 = res[:, :, 1].sum()
        expected2 = 399936.70287099993 * 2
        self.assertAlmostEqual(cross_sum2, expected2)

    def test_nearest_neighbour_multi_preproc(self):
        data1 = numpy.fromfunction(lambda y, x: y * x * 10 ** -6, (3712, 3712))
        data2 = numpy.fromfunction(
            lambda y, x: y * x * 10 ** -6, (3712, 3712)) * 2
        data = numpy.dstack((data1, data2))
        msg_con = image.ImageContainer(data, self.msg_area)
        # area_con = msg_con.resample_area_nearest_neighbour(self.area_def,
        # 50000)
        row_indices, col_indices = \
            utils.generate_nearest_neighbour_linesample_arrays(self.msg_area,
                                                               self.area_def,
                                                               50000)
        res = msg_con.get_array_from_linesample(row_indices, col_indices)
        cross_sum1 = res[:, :, 0].sum()
        expected1 = 399936.70287099993
        self.assertAlmostEqual(cross_sum1, expected1)

        cross_sum2 = res[:, :, 1].sum()
        expected2 = 399936.70287099993 * 2
        self.assertAlmostEqual(cross_sum2, expected2)

    def test_nearest_swath(self):
        data = numpy.fromfunction(lambda y, x: y * x, (50, 10))
        lons = numpy.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = numpy.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        swath_con = image.ImageContainerNearest(
            data, swath_def, 50000, segments=1)
        area_con = swath_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum = res.sum()
        expected = 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_nearest_swath_segments(self):
        data = numpy.fromfunction(lambda y, x: y * x, (50, 10))
        data = numpy.dstack(3 * (data,))
        lons = numpy.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = numpy.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        swath_con = image.ImageContainerNearest(
            data, swath_def, 50000, segments=2)
        area_con = swath_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum = res.sum()
        expected = 3 * 15874591.0
        self.assertEqual(cross_sum, expected)

    def test_bilinear(self):
        data = numpy.fromfunction(lambda y, x: y * x * 10 ** -6, (928, 928))
        msg_con = image.ImageContainerBilinear(data, self.msg_area_resize,
                                               50000, segments=1,
                                               neighbours=8)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum = res.sum()
        expected = 24712.589910252744
        self.assertAlmostEqual(cross_sum, expected)

    def test_bilinear_multi(self):
        data1 = numpy.fromfunction(lambda y, x: y * x * 10 ** -6, (928, 928))
        data2 = numpy.fromfunction(lambda y, x: y * x * 10 ** -6,
                                   (928, 928)) * 2
        data = numpy.dstack((data1, data2))
        msg_con = image.ImageContainerBilinear(data, self.msg_area_resize,
                                               50000, segments=1,
                                               neighbours=8)
        area_con = msg_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum1 = res[:, :, 0].sum()
        expected1 = 24712.589910252744
        self.assertAlmostEqual(cross_sum1, expected1)
        cross_sum2 = res[:, :, 1].sum()
        expected2 = 24712.589910252744 * 2
        self.assertAlmostEqual(cross_sum2, expected2)

    def test_bilinear_swath(self):
        data = numpy.fromfunction(lambda y, x: y * x, (50, 10))
        lons = numpy.fromfunction(lambda y, x: 3 + x, (50, 10))
        lats = numpy.fromfunction(lambda y, x: 75 - y, (50, 10))
        swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
        swath_con = image.ImageContainerBilinear(data, swath_def, 500000,
                                                 segments=1, neighbours=8)
        area_con = swath_con.resample(self.area_def)
        res = area_con.image_data
        cross_sum = res.sum()
        expected = 16852120.789503865
        self.assertAlmostEqual(cross_sum, expected, places=5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Miscellaneous utility functions for pyresample."""
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

import warnings

import numpy as np

from .cf import load_cf_area  # noqa
from .proj4 import (  # noqa
    convert_proj_floats,
    get_geostationary_height,
    proj4_dict_to_str,
    proj4_radius_parameters,
    proj4_str_to_dict,
)
from .rasterio import get_area_def_from_raster  # noqa


def get_area_def(*args, **kwargs):
    """Get an area definition."""
    from pyresample.area_config import get_area_def
    warnings.warn("'get_area_def' has moved, import it with 'from pyresample import get_area_def'", stacklevel=2)
    return get_area_def(*args, **kwargs)


def create_area_def(*args, **kwargs):
    """Create an area definition."""
    from pyresample.area_config import create_area_def
    warnings.warn("'create_area_def' has moved, import it with 'from pyresample import create_area_def'", stacklevel=2)
    return create_area_def(*args, **kwargs)


def load_area(*args, **kwargs):
    """Load an area."""
    from pyresample.area_config import load_area
    warnings.warn("'load_area' has moved, import it with 'from pyresample import load_area'", stacklevel=2)
    return load_area(*args, **kwargs)


def convert_def_to_yaml(*args, **kwargs):
    """Convert an area definition to yaml representation."""
    from pyresample.area_config import convert_def_to_yaml
    warnings.warn("'convert_def_to_yaml' has moved, import it with 'from pyresample import convert_def_to_yaml'",
                  stacklevel=2)
    return convert_def_to_yaml(*args, **kwargs)


def parse_area_file(*args, **kwargs):
    """Parse an area file."""
    from pyresample.area_config import parse_area_file
    warnings.warn("'parse_area_file' has moved, import it with 'from pyresample import parse_area_file'", stacklevel=2)
    return parse_area_file(*args, **kwargs)


def generate_quick_linesample_arrays(source_area_def, target_area_def, nprocs=1):
    """Generate linesample arrays for quick grid resampling.

    Parameters
    -----------
    source_area_def : object
        Source area definition as geometry definition object
    target_area_def : object
        Target area definition as geometry definition object
    nprocs : int, optional
        Number of processor cores to be used

    Returns
    -------
    (row_indices, col_indices) : tuple of numpy arrays
    """
    from pyresample.grid import get_linesample
    lons, lats = target_area_def.get_lonlats(nprocs)

    source_pixel_y, source_pixel_x = get_linesample(lons, lats,
                                                    source_area_def,
                                                    nprocs=nprocs)

    source_pixel_x = _downcast_index_array(source_pixel_x,
                                           source_area_def.shape[1])
    source_pixel_y = _downcast_index_array(source_pixel_y,
                                           source_area_def.shape[0])

    return source_pixel_y, source_pixel_x


def generate_nearest_neighbour_linesample_arrays(source_area_def,
                                                 target_area_def,
                                                 radius_of_influence,
                                                 nprocs=1):
    """Generate linesample arrays for nearest neighbour grid resampling.

    Parameters
    -----------
    source_area_def : object
        Source area definition as geometry definition object
    target_area_def : object
        Target area definition as geometry definition object
    radius_of_influence : float
        Cut off distance in meters
    nprocs : int, optional
        Number of processor cores to be used

    Returns
    -------
    (row_indices, col_indices) : tuple of numpy arrays
    """
    from pyresample.kd_tree import get_neighbour_info
    valid_input_index, valid_output_index, index_array, distance_array = \
        get_neighbour_info(source_area_def,
                           target_area_def,
                           radius_of_influence,
                           neighbours=1,
                           nprocs=nprocs)
    # Enumerate rows and cols
    rows = np.fromfunction(lambda i, j: i, source_area_def.shape,
                           dtype=np.int32).ravel()
    cols = np.fromfunction(lambda i, j: j, source_area_def.shape,
                           dtype=np.int32).ravel()

    # Reduce to match resampling data set
    rows_valid = rows[valid_input_index]
    cols_valid = cols[valid_input_index]

    # Get result using array indexing
    number_of_valid_points = valid_input_index.sum()
    index_mask = (index_array == number_of_valid_points)
    index_array[index_mask] = 0
    row_sample = rows_valid[index_array]
    col_sample = cols_valid[index_array]
    row_sample[index_mask] = -1
    col_sample[index_mask] = -1

    # Reshape to correct shape
    row_indices = row_sample.reshape(target_area_def.shape)
    col_indices = col_sample.reshape(target_area_def.shape)

    row_indices = _downcast_index_array(row_indices,
                                        source_area_def.shape[0])
    col_indices = _downcast_index_array(col_indices,
                                        source_area_def.shape[1])

    return row_indices, col_indices


def fwhm2sigma(fwhm):
    """Calculate sigma for gauss function from FWHM (3 dB level).

    Parameters
    ----------
    fwhm : float
        FWHM of gauss function (3 dB level of beam footprint)

    Returns
    -------
    sigma : float
        sigma for use in resampling gauss function
    """
    return fwhm / (2 * np.sqrt(np.log(2)))


def _downcast_index_array(index_array, size):
    """Try to downcast array to uint16."""
    if size <= np.iinfo(np.uint16).max:
        mask = (index_array < 0) | (index_array >= size)
        index_array[mask] = size
        index_array = index_array.astype(np.uint16)
    return index_array


def wrap_longitudes(lons):
    """Wrap longitudes to the [-180:+180[ validity range (preserves dtype).

    Parameters
    ----------
    lons : numpy array
        Longitudes in degrees

    Returns
    -------
    lons : numpy array
        Longitudes wrapped into [-180:+180[ validity range
    """
    return (lons + 180) % 360 - 180


def check_and_wrap(lons, lats):
    """Wrap longitude to [-180:+180[ and check latitude for validity.

    Args:
        lons (ndarray): Longitude degrees
        lats (ndarray): Latitude degrees

    Returns:
        lons, lats: Longitude degrees in the range [-180:180[ and the original
                    latitude array

    Raises:
        ValueError: If latitude array is not between -90 and 90
    """
    # check the latitutes
    if lats.min() < -90. or lats.max() > 90.:
        raise ValueError(
            'Some latitudes are outside the [-90.:+90] validity range')

    # check the longitudes
    if lons.min() < -180. or lons.max() >= 180.:
        # wrap longitudes to [-180;+180[
        lons = wrap_longitudes(lons)

    return lons, lats


def recursive_dict_update(d, u):
    """Update dictionary recursively.

    Copied from:

        http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = recursive_dict_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def check_slice_orientation(sli):
    """Check that the slice is slicing the right way."""
    if sli.start > sli.stop:
        if sli.step is None or sli.step > 0:
            step = -(sli.step or 1)
            sli = slice(sli.start, sli.stop, step)

    return sli



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016

# Author(s):

#   David Hoese <david.hoese@ssec.wisc.edu>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Code for resampling using the Elliptical Weighted Averaging (EWA) algorithm.

The logic and original code for this algorithm were translated from the
software package "MODIS Swath 2 Grid Toolbox" or "ms2gt" created by the
NASA National Snow & Ice Data Center (NSIDC):

    https://nsidc.org/data/modis/ms2gt/index.html

Since the project has slowed down, Terry Haran has maintained the package
and made updates available:

    http://cires1.colorado.edu/~tharan/ms2gt/

The ms2gt C executables "ll2cr" and "fornav" were rewritten for the
Polar2Grid software package created by the Space Science Engineering Center
(SSEC)/Cooperative Institute for Meteorological Satellite Studies. They were
rewritten as a combination of C++ and Cython to make them more python friendly
by David Hoese and were then copied and modified here in pyresample. The
rewrite of "ll2cr" also included an important switch from using the "mapx"
library to using the more popular and capable pyproj (PROJ.4) library.

The EWA algorithm consists of two parts "ll2cr" and "fornav" and are described
below.

ll2cr
-----

The "ll2cr" process is the first step in the EWA algorithm. It stands for
"latitude/longitude to column/row". Its main purpose is to convert
input longitude and latitude coordinates to column and row coordinates
of the destination grid. These coordinates are then used in the next step
"fornav".

fornav
------

The "fornav" or "Forward Navigation" step of the EWA algorithm is where
the actual Elliptical Weighted Averaging algorithm is run. The algorithm
maps input swath pixels to output grid pixels by averaging multiple input
pixels based on an elliptical region and other coefficients, some of which
are determined at run time.

For more information on these steps see the documentation for the
corresponding modules.
"""

from .ewa import fornav, ll2cr  # noqa

try:
    from ._legacy_dask_ewa import LegacyDaskEWAResampler  # noqa
    from .dask_ewa import DaskEWAResampler  # noqa
except ImportError:
    # dask is required but not installed
    # fallback to old numpy-only implementation (ll2cr, fornav above)
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The Boundary classes."""

import logging

import numpy as np

from pyresample.spherical import SphPolygon

logger = logging.getLogger(__name__)


class Boundary(object):
    """Boundary objects."""

    def __init__(self, lons=None, lats=None, frequency=1):
        self._contour_poly = None
        if lons is not None:
            self.lons = lons[::frequency]
        if lats is not None:
            self.lats = lats[::frequency]

    def contour(self):
        """Get lon/lats of the contour."""
        return self.lons, self.lats

    @property
    def contour_poly(self):
        """Get the Spherical polygon corresponding to the Boundary."""
        if self._contour_poly is None:
            self._contour_poly = SphPolygon(
                np.deg2rad(np.vstack(self.contour()).T))
        return self._contour_poly

    def draw(self, mapper, options, **more_options):
        """Draw the current boundary on the *mapper*."""
        self.contour_poly.draw(mapper, options, **more_options)


class AreaBoundary(Boundary):
    """Area boundary objects."""

    def __init__(self, *sides):
        Boundary.__init__(self)
        self.sides_lons, self.sides_lats = zip(*sides)
        self.sides_lons = list(self.sides_lons)
        self.sides_lats = list(self.sides_lats)

    def decimate(self, ratio):
        """Remove some points in the boundaries, but never the corners."""
        for i in range(len(self.sides_lons)):
            length = len(self.sides_lons[i])
            start = int((length % ratio) / 2)
            points = np.concatenate(([0], np.arange(start, length, ratio),
                                     [length - 1]))
            if points[1] == 0:
                points = points[1:]
            if points[-2] == (length - 1):
                points = points[:-1]
            self.sides_lons[i] = self.sides_lons[i][points]
            self.sides_lats[i] = self.sides_lats[i][points]

    def contour(self):
        """Get the (lons, lats) tuple of the boundary object."""
        lons = np.concatenate([lns[:-1] for lns in self.sides_lons])
        lats = np.concatenate([lts[:-1] for lts in self.sides_lats])

        return lons, lats


class AreaDefBoundary(AreaBoundary):
    """Boundaries for area definitions (pyresample)."""

    def __init__(self, area, frequency=1):
        lons, lats = area.get_bbox_lonlats()
        AreaBoundary.__init__(self,
                              *zip(lons, lats))

        if frequency != 1:
            self.decimate(frequency)


class SimpleBoundary(object):
    """Container for geometry boundary.

    Labelling starts in upper left corner and proceeds clockwise
    """

    def __init__(self, side1, side2, side3, side4):
        self.side1 = side1
        self.side2 = side2
        self.side3 = side3
        self.side4 = side4



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2020 Pyresample developers.
#
# This file is part of Pyresample
#
# Author(s):
#
#   Panu Lahtinen <panu.lahtinen@fmi.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""XArray version of bilinear interpolation."""

import warnings

import dask.array as da
import numpy as np
import zarr
from dask import delayed
from xarray import DataArray, Dataset

from pyresample import CHUNK_SIZE
from pyresample._spatial_mp import Proj
from pyresample.bilinear._base import (
    BilinearBase,
    array_slice_for_multiple_arrays,
    find_indices_outside_min_and_max,
    get_slicer,
    get_valid_indices_from_lonlat_boundaries,
    is_swath_to_grid_or_grid_to_grid,
    mask_coordinates,
)
from pyresample.future.resamplers._transform_utils import lonlat2xyz

CACHE_INDICES = ['bilinear_s',
                 'bilinear_t',
                 'slices_x',
                 'slices_y',
                 'mask_slices',
                 'out_coords_x',
                 'out_coords_y']

BIL_COORDINATES = {'bilinear_s': ('x1', ),
                   'bilinear_t': ('x1', ),
                   'slices_x': ('x1', 'n'),
                   'slices_y': ('x1', 'n'),
                   'mask_slices': ('x1', 'n'),
                   'out_coords_x': ('x2', ),
                   'out_coords_y': ('y2', )}


class XArrayBilinearResampler(BilinearBase):
    """Bilinear interpolation using XArray."""

    def resample(self, data, fill_value=None, nprocs=1):
        """Resample the given data."""
        del nprocs
        self.get_bil_info()
        return self.get_sample_from_bil_info(data, fill_value=fill_value, output_shape=None)

    def _create_empty_bil_info(self):
        """Create dummy info for empty result set."""
        self._valid_input_index = da.ones(self._source_geo_def.size, dtype=bool)
        self._index_array = da.ones((self._target_geo_def.size, 4), dtype=np.int32)
        self.bilinear_s = np.nan * da.zeros(self._target_geo_def.size)
        self.bilinear_t = np.nan * da.zeros(self._target_geo_def.size)
        self.slices_x = da.zeros((self._target_geo_def.size, 4), dtype=np.int32)
        self.slices_y = da.zeros((self._target_geo_def.size, 4), dtype=np.int32)
        self.out_coords_x, self.out_coords_y = self._target_geo_def.get_proj_vectors(chunks=CHUNK_SIZE)
        self.mask_slices = self._index_array >= self._source_geo_def.size

    def _get_input_xy(self):
        return _get_input_xy(self._source_geo_def,
                             Proj(self._target_geo_def.proj_str),
                             self._valid_input_index, self._index_array)

    def _get_output_xy(self):
        return _get_output_xy(self._target_geo_def)

    def _limit_output_values_to_input(self, data, res, fill_value):
        epsilon = 1e-6
        data_min = da.nanmin(data) - epsilon
        data_max = da.nanmax(data) + epsilon

        res = da.where(
            find_indices_outside_min_and_max(res, data_min, data_max),
            fill_value, res)

        return da.where(np.isnan(res), fill_value, res)

    def _reshape_to_target_area(self, res, ndim):
        shp = self._target_geo_def.shape
        if ndim == 3:
            res = da.reshape(res, (res.shape[0], shp[0], shp[1]))
        else:
            res = da.reshape(res, (shp[0], shp[1]))

        return res

    def _finalize_output_data(self, data, res, fill_value):
        res = self._limit_output_values_to_input(data, res, fill_value)
        res = self._reshape_to_target_area(res, data.ndim)

        self._add_missing_coordinates(data)

        return DataArray(res, dims=data.dims, coords=self._out_coords)

    def _add_missing_coordinates(self, data):
        self._add_x_and_y_coordinates()
        for _, dim in enumerate(data.dims):
            if dim not in self._out_coords:
                try:
                    self._out_coords[dim] = data.coords[dim]
                except KeyError:
                    pass
        self._adjust_bands_coordinates_to_match_data(data.coords)

    def _add_x_and_y_coordinates(self):
        if self._out_coords['x'] is None and self.out_coords_x is not None:
            self._out_coords['x'], self._out_coords['y'] = da.compute(self.out_coords_x, self.out_coords_y)

    def _adjust_bands_coordinates_to_match_data(self, data_coords):
        if 'bands' in data_coords:
            self._out_coords['bands'] = data_coords['bands']
        elif 'bands' in self._out_coords:
            del self._out_coords['bands']

    def _slice_data(self, data, fill_value):
        def from_delayed(delayeds, shp):
            return [da.from_delayed(d, shp, np.float32) for d in delayeds]

        if data.ndim == 2:
            shp = self.bilinear_s.shape
        else:
            shp = (data.shape[0],) + self.bilinear_s.shape

        slicer = get_slicer(data)

        return from_delayed(
            self._delayed_slice_data(
                slicer, data, fill_value), shp)

    @delayed(nout=4)
    def _delayed_slice_data(self, slicer, data, fill_value):
        return slicer(data.values, self.slices_x, self.slices_y, self.mask_slices, fill_value)

    def _get_target_proj_vectors(self):
        try:
            self.out_coords_x, self.out_coords_y = self._target_geo_def.get_proj_vectors(chunks=CHUNK_SIZE)
        except AttributeError:
            pass

    def _get_valid_input_index_and_input_coords(self):
        valid_input_index, source_lons, source_lats = \
            _get_valid_input_index(self._source_geo_def,
                                   self._target_geo_def,
                                   self._reduce_data,
                                   self._radius_of_influence)
        input_coords = lonlat2xyz(source_lons, source_lats)
        valid_input_index = np.ravel(valid_input_index)
        input_coords = input_coords[valid_input_index, :].astype(np.float64)

        return da.compute(valid_input_index, input_coords)

    def save_resampling_info(self, filename):
        """Save bilinear resampling look-up tables."""
        zarr_out = Dataset()
        for idx_name, coord in BIL_COORDINATES.items():
            var = getattr(self, idx_name)
            if isinstance(var, np.ndarray):
                var = da.from_array(var, chunks=CHUNK_SIZE)
            else:
                var = var.rechunk(CHUNK_SIZE)
            zarr_out[idx_name] = (coord, var)
        zarr_out.to_zarr(filename)

    def load_resampling_info(self, filename):
        """Load bilinear resampling look-up tables and initialize the resampler."""
        try:
            fid = zarr.open(filename, 'r')
            for val in BIL_COORDINATES:
                cache = da.array(fid[val])
                setattr(self, val, cache)
        except ValueError:
            raise IOError


def _get_output_xy(target_geo_def):
    out_x, out_y = target_geo_def.get_proj_coords(chunks=CHUNK_SIZE)
    return da.compute(np.ravel(out_x), np.ravel(out_y))


def _get_input_xy(source_geo_def, proj, valid_input_index, index_array):
    """Get x/y coordinates for the input area and reduce the data."""
    input_xy_coordinates = da.compute(*mask_coordinates(*_get_raveled_lonlats(source_geo_def)))
    valid_xy_coordinates = array_slice_for_multiple_arrays(valid_input_index, input_xy_coordinates)
    # Expand input coordinates for each output location
    expanded_coordinates = array_slice_for_multiple_arrays(index_array, valid_xy_coordinates)

    return proj(*expanded_coordinates)


def _get_raveled_lonlats(geo_def):
    lons, lats = geo_def.get_lonlats(chunks=CHUNK_SIZE)
    if lons.size == 0 or lats.size == 0:
        raise ValueError('Cannot resample empty data set')
    elif lons.size != lats.size or lons.shape != lats.shape:
        raise ValueError('Mismatch between lons and lats')

    return da.ravel(lons), da.ravel(lats)


def _get_valid_input_index(source_geo_def,
                           target_geo_def,
                           reduce_data,
                           radius_of_influence):
    """Find indices of reduced input data."""
    source_lons, source_lats = _get_raveled_lonlats(source_geo_def)

    valid_input_index = da.invert(
        find_indices_outside_min_and_max(source_lons, -180., 180.) |
        find_indices_outside_min_and_max(source_lats, -90., 90.))

    if reduce_data and is_swath_to_grid_or_grid_to_grid(source_geo_def, target_geo_def):
        valid_input_index &= get_valid_indices_from_lonlat_boundaries(
            target_geo_def, source_lons, source_lats, radius_of_influence)

    return valid_input_index, source_lons, source_lats


class XArrayResamplerBilinear(XArrayBilinearResampler):
    """Wrapper for the old resampler class."""

    def __init__(self, source_geo_def,
                 target_geo_def,
                 radius_of_influence,
                 **kwargs):
        """Initialize resampler."""
        warnings.warn("Use of XArrayResamplerBilinear is deprecated, use XArrayBilinearResampler instead")

        super(XArrayResamplerBilinear, self).__init__(
            source_geo_def,
            target_geo_def,
            radius_of_influence,
            **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test the EWA ll2cr code."""

import logging
import unittest

import numpy as np

from pyresample.test.utils import create_test_latitude, create_test_longitude

LOG = logging.getLogger(__name__)


dynamic_wgs84 = {
    "grid_name": "test_wgs84_fit",
    "origin_x": None,
    "origin_y": None,
    "width": None,
    "height": None,
    "cell_width": 0.0057,
    "cell_height": -0.0057,
    "proj4_definition": "+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs",
}

static_lcc = {
    "grid_name": "test_lcc",
    "origin_x": -1950510.636800,
    "origin_y": 4368587.226913,
    "width": 5120,
    "height": 5120,
    "cell_width": 1015.9,
    "cell_height": -1015.9,
    "proj4_definition": "+proj=lcc +a=6371200 +b=6371200 +lat_0=25 +lat_1=25 +lon_0=-95 +units=m +no_defs",
}

static_geo_whole_earth = {
    "grid_name": "test_geo",
    "origin_x": -180.0,
    "origin_y": 40,
    "width": 361,
    "height": 181,
    "cell_width": 1.0,
    "cell_height": 1.0,
    "proj4_definition": "+proj=longlat +datum=WGS84 +no_defs +type=crs",
}


class TestLL2CRStatic(unittest.TestCase):
    """Test ll2cr when given a complete area definition."""

    def test_lcc_basic1(self):
        from pyresample.ewa import _ll2cr
        lon_arr = create_test_longitude(-95.0, -75.0, (50, 100), dtype=np.float64)
        lat_arr = create_test_latitude(18.0, 40.0, (50, 100), dtype=np.float64)
        grid_info = static_lcc.copy()
        fill_in = np.nan
        proj_str = grid_info["proj4_definition"]
        cw = grid_info["cell_width"]
        ch = grid_info["cell_height"]
        ox = grid_info["origin_x"]
        oy = grid_info["origin_y"]
        w = grid_info["width"]
        h = grid_info["height"]
        points_in_grid = _ll2cr.ll2cr_static(lon_arr, lat_arr, fill_in, proj_str,
                                             cw, ch, w, h, ox, oy)
        self.assertEqual(points_in_grid, lon_arr.size, "all these test points should fall in this grid")

    def test_geo_antimeridian(self):
        """Ensure output for anti-meridian crossing input includes all points."""
        from pyresample.ewa import _ll2cr
        lon_arr = create_test_longitude(160.0, 200.0, (50, 100), dtype=np.float64)
        lat_arr = create_test_latitude(40.0, 60.0, (50, 100), dtype=np.float64)

        # wrap values in lon_arr so -180 ≤ longitude (degrees east) ≤ 180
        lon_arr[lon_arr > 180] -= 360.0

        grid_info = static_geo_whole_earth.copy()
        fill_in = np.nan
        proj_str = grid_info['proj4_definition']
        cw = grid_info['cell_width']
        ch = grid_info['cell_height']
        ox = grid_info['origin_x']
        oy = grid_info['origin_y']
        w = grid_info['width']
        h = grid_info['height']
        points_in_grid = _ll2cr.ll2cr_static(lon_arr, lat_arr, fill_in, proj_str,
                                             cw, ch, w, h, ox, oy)
        self.assertEqual(points_in_grid, lon_arr.size,
                         'all these test points should fall in this grid')

    def test_lcc_fail1(self):
        from pyresample.ewa import _ll2cr
        lon_arr = create_test_longitude(-15.0, 15.0, (50, 100), dtype=np.float64)
        lat_arr = create_test_latitude(18.0, 40.0, (50, 100), dtype=np.float64)
        grid_info = static_lcc.copy()
        fill_in = np.nan
        proj_str = grid_info["proj4_definition"]
        cw = grid_info["cell_width"]
        ch = grid_info["cell_height"]
        ox = grid_info["origin_x"]
        oy = grid_info["origin_y"]
        w = grid_info["width"]
        h = grid_info["height"]
        points_in_grid = _ll2cr.ll2cr_static(lon_arr, lat_arr, fill_in, proj_str,
                                             cw, ch, w, h, ox, oy)
        self.assertEqual(points_in_grid, 0, "none of these test points should fall in this grid")


class TestLL2CRDynamic(unittest.TestCase):
    """Test ll2cr when given partial area definition information."""

    def test_latlong_basic1(self):
        from pyresample.ewa import _ll2cr
        lon_arr = create_test_longitude(-95.0, -75.0, (50, 100), dtype=np.float64)
        lat_arr = create_test_latitude(15.0, 30.0, (50, 100), dtype=np.float64)
        grid_info = dynamic_wgs84.copy()
        fill_in = np.nan
        proj_str = grid_info["proj4_definition"]
        cw = grid_info["cell_width"]
        ch = grid_info["cell_height"]
        ox = grid_info["origin_x"]
        oy = grid_info["origin_y"]
        w = grid_info["width"]
        h = grid_info["height"]
        points_in_grid, lon_res, lat_res, ox, oy, w, h = _ll2cr.ll2cr_dynamic(lon_arr, lat_arr, fill_in, proj_str,
                                                                              cw, ch, w, h, ox, oy)
        self.assertEqual(points_in_grid, lon_arr.size, "all points should be contained in a dynamic grid")
        self.assertIs(lon_arr, lon_res)
        self.assertIs(lat_arr, lat_res)
        self.assertEqual(lon_arr[0, 0], 0, "ll2cr returned the wrong result for a dynamic latlong grid")
        self.assertEqual(lat_arr[-1, 0], 0, "ll2cr returned the wrong result for a dynamic latlong grid")

    def test_latlong_basic2(self):
        from pyresample.ewa import _ll2cr
        lon_arr = create_test_longitude(-95.0, -75.0, (50, 100), twist_factor=0.6, dtype=np.float64)
        lat_arr = create_test_latitude(15.0, 30.0, (50, 100), twist_factor=-0.1, dtype=np.float64)
        grid_info = dynamic_wgs84.copy()
        fill_in = np.nan
        proj_str = grid_info["proj4_definition"]
        cw = grid_info["cell_width"]
        ch = grid_info["cell_height"]
        ox = grid_info["origin_x"]
        oy = grid_info["origin_y"]
        w = grid_info["width"]
        h = grid_info["height"]
        points_in_grid, lon_res, lat_res, ox, oy, w, h = _ll2cr.ll2cr_dynamic(lon_arr, lat_arr, fill_in, proj_str,
                                                                              cw, ch, w, h, ox, oy)
        self.assertEqual(points_in_grid, lon_arr.size, "all points should be contained in a dynamic grid")
        self.assertIs(lon_arr, lon_res)
        self.assertIs(lat_arr, lat_res)
        self.assertEqual(lon_arr[0, 0], 0, "ll2cr returned the wrong result for a dynamic latlong grid")
        self.assertEqual(lat_arr[-1, 0], 0, "ll2cr returned the wrong result for a dynamic latlong grid")

    def test_latlong_dateline1(self):
        from pyresample.ewa import _ll2cr
        lon_arr = create_test_longitude(165.0, -165.0, (50, 100), twist_factor=0.6, dtype=np.float64)
        lat_arr = create_test_latitude(15.0, 30.0, (50, 100), twist_factor=-0.1, dtype=np.float64)
        grid_info = dynamic_wgs84.copy()
        fill_in = np.nan
        proj_str = grid_info["proj4_definition"]
        cw = grid_info["cell_width"]
        ch = grid_info["cell_height"]
        ox = grid_info["origin_x"]
        oy = grid_info["origin_y"]
        w = grid_info["width"]
        h = grid_info["height"]
        points_in_grid, lon_res, lat_res, ox, oy, w, h = _ll2cr.ll2cr_dynamic(lon_arr, lat_arr, fill_in, proj_str,
                                                                              cw, ch, w, h, ox, oy)
        self.assertEqual(points_in_grid, lon_arr.size, "all points should be contained in a dynamic grid")
        self.assertIs(lon_arr, lon_res)
        self.assertIs(lat_arr, lat_res)
        self.assertEqual(lon_arr[0, 0], 0, "ll2cr returned the wrong result for a dynamic latlong grid")
        self.assertEqual(lat_arr[-1, 0], 0, "ll2cr returned the wrong result for a dynamic latlong grid")
        self.assertTrue(np.all(np.diff(lon_arr[0]) >= 0), "ll2cr didn't return monotonic columns over the dateline")


class TestLL2CRWrapper(unittest.TestCase):
    """Test ll2cr high-level python wrapper."""

    def test_basic1(self):
        from pyresample.ewa import ll2cr
        from pyresample.geometry import AreaDefinition, SwathDefinition
        from pyresample.utils import proj4_str_to_dict
        lon_arr = create_test_longitude(-95.0, -75.0, (50, 100), dtype=np.float64)
        lat_arr = create_test_latitude(18.0, 40.0, (50, 100), dtype=np.float64)
        swath_def = SwathDefinition(lon_arr, lat_arr)
        grid_info = static_lcc.copy()
        cw = grid_info["cell_width"]
        ch = grid_info["cell_height"]
        ox = grid_info["origin_x"]
        oy = grid_info["origin_y"]
        w = grid_info["width"]
        h = grid_info["height"]
        half_w = abs(cw / 2.)
        half_h = abs(ch / 2.)
        extents = [
            ox - half_w, oy - h * abs(ch) - half_h,
            ox + w * abs(cw) + half_w, oy + half_h
        ]
        area = AreaDefinition('test_area', 'test_area', 'test_area',
                              proj4_str_to_dict(grid_info['proj4_definition']),
                              w, h, extents)
        points_in_grid, lon_res, lat_res, = ll2cr(swath_def, area,
                                                  fill=np.nan, copy=False)
        self.assertEqual(points_in_grid, lon_arr.size, "all points should be contained in a dynamic grid")
        self.assertIs(lon_arr, lon_res)
        self.assertIs(lat_arr, lat_res)
        self.assertEqual(points_in_grid, lon_arr.size, "all these test points should fall in this grid")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2013 - 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Some generalized spherical functions.

base type is a numpy array of size (n, 2) (2 for lon and lats)
"""

import logging

import numpy as np

logger = logging.getLogger(__name__)


class SCoordinate(object):
    """Spherical coordinates.

    The ``lon`` and ``lat`` coordinates should be provided in radians.

    """

    def __init__(self, lon, lat):
        self.lon = lon
        self.lat = lat

    def cross2cart(self, point):
        """Compute the cross product, and convert to cartesian coordinates."""
        lat1 = self.lat
        lon1 = self.lon
        lat2 = point.lat
        lon2 = point.lon

        ad = np.sin(lat1 - lat2) * np.cos((lon1 - lon2) / 2.0)
        be = np.sin(lat1 + lat2) * np.sin((lon1 - lon2) / 2.0)
        c = np.sin((lon1 + lon2) / 2.0)
        f = np.cos((lon1 + lon2) / 2.0)
        g = np.cos(lat1)
        h = np.cos(lat2)
        i = np.sin(lon2 - lon1)
        res = CCoordinate(np.array([-ad * c + be * f,
                                    ad * f + be * c,
                                    g * h * i]))

        return res

    def to_cart(self):
        """Convert to cartesian."""
        return CCoordinate(np.array([np.cos(self.lat) * np.cos(self.lon),
                                     np.cos(self.lat) * np.sin(self.lon),
                                     np.sin(self.lat)]))

    def distance(self, point):
        """Get distance using Vincenty formula."""
        dlambda = self.lon - point.lon
        num = ((np.cos(point.lat) * np.sin(dlambda)) ** 2 +
               (np.cos(self.lat) * np.sin(point.lat) -
                np.sin(self.lat) * np.cos(point.lat) *
                np.cos(dlambda)) ** 2)
        den = (np.sin(self.lat) * np.sin(point.lat) +
               np.cos(self.lat) * np.cos(point.lat) * np.cos(dlambda))

        return np.arctan2(num ** .5, den)

    def hdistance(self, point):
        """Get distance using Haversine formula."""
        return 2 * np.arcsin((np.sin((point.lat - self.lat) / 2.0) ** 2.0 +
                              np.cos(point.lat) * np.cos(self.lat) *
                              np.sin((point.lon - self.lon) / 2.0) ** 2.0) ** .5)

    def __ne__(self, other):
        """Check inequality."""
        return not self.__eq__(other)

    def __eq__(self, other):
        """Check equality."""
        return np.allclose((self.lon, self.lat), (other.lon, other.lat))

    def __str__(self):
        """Get simplified representation of lon/lat arrays in degrees."""
        return str((np.rad2deg(self.lon), np.rad2deg(self.lat)))

    def __repr__(self):
        """Get simplified representation of lon/lat arrays in degrees."""
        return str((np.rad2deg(self.lon), np.rad2deg(self.lat)))

    def __iter__(self):
        """Get iterator over lon/lat pairs."""
        return zip([self.lon, self.lat]).__iter__()


class CCoordinate(object):
    """Cartesian coordinates."""

    def __init__(self, cart):
        self.cart = np.array(cart)

    def norm(self):
        """Get Euclidean norm of the vector."""
        return np.sqrt(np.einsum('...i, ...i', self.cart, self.cart))

    def normalize(self):
        """Normalize the vector."""
        self.cart /= np.sqrt(np.einsum('...i, ...i', self.cart, self.cart))

        return self

    def cross(self, point):
        """Get cross product with another vector."""
        return CCoordinate(np.cross(self.cart, point.cart))

    def dot(self, point):
        """Get dot product with another vector."""
        return np.inner(self.cart, point.cart)

    def __ne__(self, other):
        """Check inequality."""
        return not self.__eq__(other)

    def __eq__(self, other):
        """Check equality."""
        return np.allclose(self.cart, other.cart)

    def __str__(self):
        """Get simplified representation."""
        return str(self.cart)

    def __repr__(self):
        """Get simplified representation."""
        return str(self.cart)

    def __add__(self, other):
        """Add."""
        try:
            return CCoordinate(self.cart + other.cart)
        except AttributeError:
            return CCoordinate(self.cart + np.array(other))

    def __radd__(self, other):
        """Add."""
        return self.__add__(other)

    def __mul__(self, other):
        """Multiply."""
        try:
            return CCoordinate(self.cart * other.cart)
        except AttributeError:
            return CCoordinate(self.cart * np.array(other))

    def __rmul__(self, other):
        """Multiply."""
        return self.__mul__(other)

    def to_spherical(self):
        """Convert to Spherical coordinate object."""
        return SCoordinate(np.arctan2(self.cart[1], self.cart[0]),
                           np.arcsin(self.cart[2]))


EPSILON = 0.0000001


def modpi(val, mod=np.pi):
    """Put *val* between -*mod* and *mod*."""
    return (val + mod) % (2 * mod) - mod


class Arc(object):
    """An arc of the great circle between two points."""

    def __init__(self, start, end):
        self.start, self.end = start, end

    def __eq__(self, other):
        """Check equality."""
        if self.start == other.start and self.end == other.end:
            return 1
        return 0

    def __ne__(self, other):
        """Check not equal comparison."""
        return not self.__eq__(other)

    def __str__(self):
        """Get simplified representation."""
        return str(self.start) + " -> " + str(self.end)

    def __repr__(self):
        """Get simplified representation."""
        return str(self.start) + " -> " + str(self.end)

    def angle(self, other_arc):
        """Oriented angle between two arcs.

        Returns:
            Angle in radians. A straight line will be 0. A clockwise path
            will be a negative angle and counter-clockwise will be positive.

        """
        if self.start == other_arc.start:
            a__ = self.start
            b__ = self.end
            c__ = other_arc.end
        elif self.start == other_arc.end:
            a__ = self.start
            b__ = self.end
            c__ = other_arc.start
        elif self.end == other_arc.end:
            a__ = self.end
            b__ = self.start
            c__ = other_arc.start
        elif self.end == other_arc.start:
            a__ = self.end
            b__ = self.start
            c__ = other_arc.end
        else:
            raise ValueError("No common point in angle computation.")

        ua_ = a__.cross2cart(b__)
        ub_ = a__.cross2cart(c__)

        val = ua_.dot(ub_) / (ua_.norm() * ub_.norm())
        if abs(val - 1) < EPSILON:
            angle = 0
        elif abs(val + 1) < EPSILON:
            angle = np.pi
        else:
            angle = np.arccos(val)

        n__ = ua_.normalize()
        if n__.dot(c__.to_cart()) > 0:
            return -angle
        else:
            return angle

    def intersections(self, other_arc):
        """Give the two intersections of the greats circles defined by the current arc and *other_arc*.

        From http://williams.best.vwh.net/intersect.htm
        """
        if self.end.lon - self.start.lon > np.pi:
            self.end.lon -= 2 * np.pi
        if other_arc.end.lon - other_arc.start.lon > np.pi:
            other_arc.end.lon -= 2 * np.pi
        if self.end.lon - self.start.lon < -np.pi:
            self.end.lon += 2 * np.pi
        if other_arc.end.lon - other_arc.start.lon < -np.pi:
            other_arc.end.lon += 2 * np.pi

        ea_ = self.start.cross2cart(self.end).normalize()
        eb_ = other_arc.start.cross2cart(other_arc.end).normalize()

        cross = ea_.cross(eb_)
        lat = np.arctan2(cross.cart[2],
                         np.sqrt(cross.cart[0] ** 2 + cross.cart[1] ** 2))
        lon = np.arctan2(cross.cart[1], cross.cart[0])

        return (SCoordinate(lon, lat),
                SCoordinate(modpi(lon + np.pi), -lat))

    def intersects(self, other_arc):
        """Check if the current arc and the *other_arc* intersect.

        An arc is defined as the shortest tracks between two points.
        """
        return bool(self.intersection(other_arc))

    def intersection(self, other_arc):
        """Return where, if the current arc and the *other_arc* intersect.

        None is returned if there is not intersection. An arc is defined
        as the shortest tracks between two points.
        """
        if self == other_arc:
            return None

        for i in self.intersections(other_arc):
            a__ = self.start
            b__ = self.end
            c__ = other_arc.start
            d__ = other_arc.end

            ab_ = a__.hdistance(b__)
            cd_ = c__.hdistance(d__)

            if(((i in (a__, b__)) or
                (abs(a__.hdistance(i) + b__.hdistance(i) - ab_) < EPSILON)) and
               ((i in (c__, d__)) or
                    (abs(c__.hdistance(i) + d__.hdistance(i) - cd_) < EPSILON))):
                return i
        return None

    def get_next_intersection(self, arcs, known_inter=None):
        """Get the next intersection between the current arc and *arcs*."""
        res = []
        for arc in arcs:
            inter = self.intersection(arc)
            if (inter is not None and
                    inter != arc.end and
                    inter != self.end):
                res.append((inter, arc))

        def dist(args):
            """Get distance key."""
            return self.start.distance(args[0])

        take_next = False
        for inter, arc in sorted(res, key=dist):
            if known_inter is not None:
                if known_inter == inter:
                    take_next = True
                elif take_next:
                    return inter, arc
            else:
                return inter, arc

        return None, None


class SphPolygon:
    """Spherical polygon.

    Represents a polygon on a spherical geoid.  Initialise with
    an ndarray of shape ``[N, 2]`` where the first column contains longitudes
    and the second column contains latitudes.  The units should be in radians.
    The inside of the polygon is defined by the vertices being defined clockwise
    around it.

    The optional second argument ``radius`` indicates the radius of the
    spherical geoid on which calculations occur.

    """

    def __init__(self, vertices, radius=1):
        """Initialise SphPolygon object.

        Args:
            vertices (np.ndarray): ndarray of shape ``[N, 2]`` with ``N``
                points describing a polygon clockwise.  First column
                describes longitudes, second column describes latitudes.  Units
                should be in radians.
            radius (optional, number): Radius of spherical planet.
        """
        self.vertices = vertices.astype(np.float64, copy=False)
        self.lon = self.vertices[:, 0]
        self.lat = self.vertices[:, 1]
        self.radius = radius
        self.cvertices = np.array([np.cos(self.lat) * np.cos(self.lon),
                                   np.cos(self.lat) * np.sin(self.lon),
                                   np.sin(self.lat)]).T * radius
        self.x__ = self.cvertices[:, 0]
        self.y__ = self.cvertices[:, 1]
        self.z__ = self.cvertices[:, 2]

    def invert(self):
        """Invert the polygon."""
        self.vertices = np.flipud(self.vertices)
        self.cvertices = np.flipud(self.cvertices)
        self.lon = self.vertices[:, 0]
        self.lat = self.vertices[:, 1]
        self.x__ = self.cvertices[:, 0]
        self.y__ = self.cvertices[:, 1]
        self.z__ = self.cvertices[:, 2]

    def inverse(self):
        """Return an inverse of the polygon."""
        return SphPolygon(np.flipud(self.vertices), radius=self.radius)

    def aedges(self):
        """Get generator over the edges, in arcs of Coordinates."""
        for (lon_start, lat_start), (lon_stop, lat_stop) in self.edges():
            yield Arc(SCoordinate(lon_start, lat_start),
                      SCoordinate(lon_stop, lat_stop))

    def edges(self):
        """Get generator over the edges, in geographical coordinates."""
        for i in range(len(self.lon) - 1):
            yield (self.lon[i], self.lat[i]), (self.lon[i + 1], self.lat[i + 1])
        yield (self.lon[i + 1], self.lat[i + 1]), (self.lon[0], self.lat[0])

    def area(self):
        """Find the area of a polygon.

        The inside of the polygon is defined by having the vertices enumerated
        clockwise around it.

        Uses the algorithm described in [bev1987]_.

        .. [bev1987] , Michael Bevis and Greg Cambareri,
           "Computing the area of a spherical polygon of arbitrary shape",
           in *Mathematical Geology*, May 1987, Volume 19, Issue 4, pp 335-346.

        Note: The article mixes up longitudes and latitudes in equation 3! Look
        at the fortran code appendix for the correct version.

        The units are the square of the radius passed to the constructor.  For
        example, to calculate the area in km² of a polygon near the equator of a
        spherical planet with a radius of 6371 km (similar to Earth):

        >>> pol = SphPolygon(np.deg2rad(np.array([[0., 0.], [0., 1.], [1., 1.], [1., 0.]])),
                             radius=6371)
        >>> print(pol.area())
        12363.997753690213

        If `SphPolygon` was constructed without passing any units, the result
        has units of square radii (i.e., the polygon containing the entire
        planet would have area 4π).
        """
        phi_a = self.lat
        phi_p = self.lat.take(np.arange(len(self.lat)) + 1, mode="wrap")
        phi_b = self.lat.take(np.arange(len(self.lat)) + 2, mode="wrap")
        lam_a = self.lon
        lam_p = self.lon.take(np.arange(len(self.lon)) + 1, mode="wrap")
        lam_b = self.lon.take(np.arange(len(self.lon)) + 2, mode="wrap")

        new_lons_a = np.arctan2(np.sin(lam_a - lam_p) * np.cos(phi_a),
                                np.sin(phi_a) * np.cos(phi_p) -
                                np.cos(phi_a) * np.sin(phi_p) *
                                np.cos(lam_a - lam_p))

        new_lons_b = np.arctan2(np.sin(lam_b - lam_p) * np.cos(phi_b),
                                np.sin(phi_b) * np.cos(phi_p) -
                                np.cos(phi_b) * np.sin(phi_p) *
                                np.cos(lam_b - lam_p))

        alpha = new_lons_a - new_lons_b
        alpha[alpha < 0] += 2 * np.pi

        return (sum(alpha) - (len(self.lon) - 2) * np.pi) * self.radius ** 2

    def _bool_oper(self, other, sign=1):
        """Perform a boolean operation on this and *other* polygons.abs.

        By default, or when sign is 1, the union is perfomed. If sign is -1,
        the intersection of the polygons is returned.

        The algorithm works this way: Find an intersection between the two
        polygons. If none can be found, then the two polygons are either not
        overlapping, or one is entirely included in the other. Otherwise,
        follow the edges of a polygon until another intersection is
        encountered, at which point you start following the edges of the other
        polygon, and so on until you come back to the first intersection. In
        which direction to follow the edges of the polygons depends if you are
        interested in the union or the intersection of the two polygons.
        """
        def rotate_arcs(start_arc, arcs):
            idx = arcs.index(start_arc)
            return arcs[idx:] + arcs[:idx]

        arcs1 = [edge for edge in self.aedges()]
        arcs2 = [edge for edge in other.aedges()]

        nodes = []

        # find the first intersection, to start from.
        for edge1 in arcs1:
            inter, edge2 = edge1.get_next_intersection(arcs2)
            if inter is not None and inter != edge1.end and inter != edge2.end:
                break

        # if no intersection is found, find out if the one poly is included in
        # the other.
        if inter is None:
            polys = [0, self, other]
            if self._is_inside(other):
                return polys[-sign]
            if other._is_inside(self):
                return polys[sign]

            return None

        # starting from the intersection, follow the edges of one of the
        # polygons.

        while True:
            arcs1 = rotate_arcs(edge1, arcs1)
            arcs2 = rotate_arcs(edge2, arcs2)

            narcs1 = arcs1 + [edge1]
            narcs2 = arcs2 + [edge2]

            arc1 = Arc(inter, edge1.end)
            arc2 = Arc(inter, edge2.end)

            if np.sign(arc1.angle(arc2)) != sign:
                arcs1, arcs2 = arcs2, arcs1
                narcs1, narcs2 = narcs2, narcs1

            nodes.append(inter)

            for edge1 in narcs1:
                inter, edge2 = edge1.get_next_intersection(narcs2, inter)
                if inter is not None:
                    break
                elif len(nodes) > 0 and edge1.end not in [nodes[-1], nodes[0]]:
                    nodes.append(edge1.end)

            if inter is None and len(nodes) > 2 and nodes[-1] == nodes[0]:
                nodes = nodes[:-1]
                break
            if inter == nodes[0]:
                break
        return SphPolygon(np.array([(node.lon, node.lat) for node in nodes]), radius=self.radius)

    def union(self, other):
        """Return the union of this and `other` polygon.

        NB! If the two polygons do not overlap (they have nothing in common) None is returned.
        """
        return self._bool_oper(other, 1)

    def intersection(self, other):
        """Return the intersection of this and `other` polygon."""
        return self._bool_oper(other, -1)

    def _is_inside(self, other):
        """Check if the polygon is entirely inside the other.

        Should be used with :meth:`inter` first to check if the is a
        known intersection.
        """
        anti_lon_0 = self.lon[0] + np.pi
        if anti_lon_0 > np.pi:
            anti_lon_0 -= np.pi * 2

        anti_lon_1 = self.lon[1] + np.pi
        if anti_lon_1 > np.pi:
            anti_lon_1 -= np.pi * 2

        arc1 = Arc(SCoordinate(self.lon[1],
                               self.lat[1]),
                   SCoordinate(anti_lon_0,
                               -self.lat[0]))

        arc2 = Arc(SCoordinate(anti_lon_0,
                               -self.lat[0]),
                   SCoordinate(anti_lon_1,
                               -self.lat[1]))

        arc3 = Arc(SCoordinate(anti_lon_1,
                               -self.lat[1]),
                   SCoordinate(self.lon[0],
                               self.lat[0]))

        other_arcs = [edge for edge in other.aedges()]
        for arc in [arc1, arc2, arc3]:
            inter, other_arc = arc.get_next_intersection(other_arcs)
            if inter is not None:
                sarc = Arc(arc.start, inter)
                earc = Arc(inter, other_arc.end)
                return sarc.angle(earc) < 0
        return other.area() > (2 * np.pi * other.radius ** 2)

    def __str__(self):
        """Get numpy representation of vertices."""
        return str(np.rad2deg(self.vertices))
